from functools import cached_property
from json import load
from pathlib import Path
from typing import Dict, List, Union, cast

from printo import repred

from skelet.sources.abstract import AbstractSource, ExpectedType


@repred(positionals=['path'], filters={'allow_non_existent_files': lambda x: x != True})  # type: ignore[arg-type]
class JSONSource(AbstractSource[ExpectedType]):
    def __init__(self, path: Union[str, Path], allow_non_existent_files: bool = True) -> None:
        self.path = path
        self.allow_non_existent_files = allow_non_existent_files

    def __getitem__(self, key: str) -> ExpectedType:
        return self.data[key]

    @cached_property
    def data(self) -> Dict[str, ExpectedType]:
        try:
            with open(self.path, 'r') as file:
                return cast(Dict[str, ExpectedType], load(file))

        except FileNotFoundError:
            if self.allow_non_existent_files:
                return {}
            raise

    @classmethod
    def for_library(cls, library_name: str) -> List['JSONSource[ExpectedType]']:
        if not library_name.isidentifier():
            raise ValueError('The library name can only be a valid Python identifier.')

        return [cls(f'{library_name}.json'), cls(f'.{library_name}.json')]



from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
    get_origin,
    get_type_hints,
)

# TODO: check, EllipsisType was added to types module in Python 3.10.
try:
    from types import EllipsisType  # type: ignore[attr-defined, unused-ignore]
except ImportError:  # pragma: no cover
    EllipsisType = type(...)  # type: ignore[misc, unused-ignore]

from collections.abc import Sequence
from sys import version_info
from threading import Lock

from denial import InnerNoneType
from locklib import ContextLockProtocol
from sigmatch import PossibleCallMatcher, SignatureMismatchError
from simtypes import check

from skelet.sources.abstract import AbstractSource, ExpectedType
from skelet.sources.collection import SourcesCollection
from skelet.storage import Storage

ValueType = TypeVar('ValueType')

if version_info < (3, 9):  # pragma: no cover
    SequenceWithStrings = Sequence
else:  # pragma: no cover
    SequenceWithStrings = Sequence[str]  # type: ignore[misc, assignment]

sentinel = InnerNoneType()

class Field(Generic[ValueType]):
    def __init__(  # noqa: PLR0913, PLR0915
        self,
        default: Union[ValueType, InnerNoneType] = sentinel,
        /,
        default_factory: Optional[Callable[[], ValueType]] = None,
        doc: Optional[str] = None,
        alias: Optional[str] = None,
        sources: Optional[List[Union[AbstractSource[ExpectedType], EllipsisType]]] = None,
        read_only: bool = False,
        validation: Optional[Union[Dict[str, Callable[[ValueType], bool]], Callable[[ValueType], bool]]] = None,
        validate_default: bool = True,
        secret: bool = False,
        action: Optional[Callable[[ValueType, ValueType, Storage], Any]] = None,
        read_lock: bool = False,
        conflicts: Optional[Dict[str, Callable[[ValueType, ValueType, Any, Any], bool]]] = None,
        reverse_conflicts: bool = True,
        conversion: Optional[Callable[[ValueType], ValueType]] = None,
        share_mutex_with: Optional[SequenceWithStrings] = None,  # type: ignore[type-arg]
    ) -> None:
        if default_factory is not None:
            if default is not sentinel:
                raise ValueError('You can define a default value or a factory for default values, but not all at the same time.')
            if not PossibleCallMatcher().match(default_factory):
                raise SignatureMismatchError('The default value factory should not expect any arguments.')

        if validation is not None:
            validation_matcher = PossibleCallMatcher('.')
            if isinstance(validation, dict):
                for validator_message, validator in validation.items():
                    if not validation_matcher.match(validator):
                        raise SignatureMismatchError(f'Field validator with message {validator_message!r} is incorrect: a function that accepts only one positional argument is expected.')
            elif not validation_matcher.match(validation):
                raise SignatureMismatchError('A function that accepts only one positional argument is expected as a field validator.')

        if action is not None and not PossibleCallMatcher('...').match(action):
            raise SignatureMismatchError('The callback for each field change must take 3 arguments: the old field value, the new value, and the storage object itself.')

        if conflicts is not None:
            conflict_checker_matcher = PossibleCallMatcher('....')
            for potentially_conflicting_field_name, conflict_checker in conflicts.items():
                if not conflict_checker_matcher.match(conflict_checker):
                    raise SignatureMismatchError(f'The function for checking conflicts with field {potentially_conflicting_field_name!r} is bad; it should take four positional arguments: the old value of this field, the new value of this field, the old value of the conflicting field, and the new value of the conflicting field (for reverse checks).')

        if conversion is not None:
            if not PossibleCallMatcher('.').match(conversion):
                raise SignatureMismatchError('The value converter must accept only one argument: the value before conversion.')
            if default is not sentinel:
                self._default_before_conversion: Union[ValueType, InnerNoneType] = default
                self._default: Union[ValueType, InnerNoneType] = conversion(cast(ValueType, default))
            else:
                self._default_before_conversion = sentinel
                self._default = default
        else:
            self._default_before_conversion = sentinel
            self._default = default

        self._default_factory = default_factory
        self.read_only = read_only
        self.doc = doc
        self.alias = alias
        self.sources = sources
        self.validation = validation
        self.validate_default = validate_default
        self.secret = secret
        self.change_action = action
        self.conflicts = conflicts
        self.reverse_conflicts_on = reverse_conflicts
        self.conversion = conversion
        self.share_mutex_with = share_mutex_with

        self.name: Optional[str] = None
        self.base_class: Optional[Type[Storage]] = None
        self.exception: Optional[BaseException] = None
        self.type_hint = Any

        self.lock: ContextLockProtocol = Lock()

        if read_lock:
            self.real_get = self.locked_get  # type: ignore[method-assign]
        else:
            self.real_get = self.unlocked_get  # type: ignore[method-assign]

    def __set_name__(self, owner: Type[Storage], name: str) -> None:
        if name.startswith('_'):
            self.raise_exception_in_storage(ValueError(f'Field name "{name}" cannot start with an underscore.'), raising_on=False)

        with self.lock:
            self.name = name
            if self.alias is None:
                self.alias = self.name
            self.type_hint = get_type_hints(owner).get(name, Any)

            if self.base_class is not None:
                self.raise_exception_in_storage(TypeError(f'{self.get_field_name_representation()} cannot be used in {owner.__name__} because it is already used in {self.base_class.__name__}.'), raising_on=False)

            if not issubclass(owner, Storage):
                self.raise_exception_in_storage(TypeError(f'{self.get_field_name_representation()} can only be used in Storage subclasses.'), raising_on=True)

            self.base_class = owner

            if self._default_before_conversion is not sentinel:
                self.check_type_hints(cast(ValueType, self._default_before_conversion))

            self.set_field_names(owner, name)
            if self._default is not sentinel:
                self.check_type_hints(cast(ValueType, self._default))
                if self.validate_default:
                    self.check_value(cast(ValueType, self._default))

    def __get__(self, instance: Storage, instance_class: Type[Storage]) -> ValueType:
        if instance is None:
            return self

        return self.real_get(instance, instance_class)

    def real_get(self, instance: Storage, instance_class: Type[Storage]) -> ValueType:
        raise NotImplementedError('If you see this error, it means something is broken.')  # pragma: no cover

    def locked_get(self, instance: Storage, instance_class: Type[Storage]) -> ValueType:
        with self.get_field_lock(instance):
            return self.unlocked_get(instance, instance_class)

    def unlocked_get(self, instance: Storage, instance_class: Type[Storage]) -> ValueType:  #noqa: ARG002
        return cast(ValueType, instance.__values__.get(cast(str, self.name)))

    def __set__(self, instance: Storage, value: ValueType) -> None:
        if self.read_only:
            raise AttributeError(f'{self.get_field_name_representation()} is read-only.')

        self.check_type_hints(value, raise_all=True)

        if self.conversion is not None:
            value = self.conversion(value)
            self.check_type_hints(value, raise_all=True)

        self.check_value(value, raise_all=True)

        with self.get_field_lock(instance):
            old_value = self.unlocked_get(instance, type(instance))
            if self.conflicts is not None:
                for other_field_name, checker in self.conflicts.items():
                    other_field = getattr(type(instance), other_field_name)
                    other_field_value = other_field.unlocked_get(instance, type(instance))
                    if checker(old_value, value, other_field_value, other_field_value):
                        raise ValueError(f'The new {self.get_value_representation(value)} value of the {self.get_field_name_representation()} conflicts with the {other_field.get_value_representation(other_field_value)} value of the {other_field.get_field_name_representation()}.')

            if self.name in instance.__reverse_conflicts__:
                for other_field_name in instance.__reverse_conflicts__[self.name]:
                    other_field = getattr(type(instance), other_field_name)
                    other_field_value = other_field.unlocked_get(instance, type(instance))
                    other_field_checker = other_field.conflicts[self.name]
                    if other_field_checker(other_field_value, other_field_value, old_value, value):
                        raise ValueError(f'The new {self.get_value_representation(value)} value of the {self.get_field_name_representation()} conflicts with the {other_field.get_value_representation(other_field_value)} value of the {other_field.get_field_name_representation()}.')

            instance.__values__[cast(str, self.name)] = value
            if self.change_action is not None and value != old_value:
                self.change_action(old_value, value, instance)

    def __delete__(self, instance: Any) -> None:
        raise AttributeError(f"You can't delete the {self.get_field_name_representation()} value.")

    def set_field_names(self, owner: Type[Storage], name: str) -> None:
        if '__field_names__' not in owner.__dict__:
            owner.__field_names__ = []
            known_names = set()
            for parent in owner.__mro__:  # pragma: no branch
                if parent is owner:
                    continue
                if parent is Storage:
                    break
                for field_name in cast(Storage, parent).__field_names__:
                    if field_name not in known_names:
                        known_names.add(field_name)
                        owner.__field_names__.append(field_name)
        else:
            known_names = set(owner.__field_names__)

        if name not in known_names:  # pragma: no branch
            cast(List[str], owner.__field_names__).append(name)

    def check_type_hints(self, value: ValueType, strict: bool = False, raise_all: bool = False) -> None:
        if not check(value, self.type_hint, strict=strict):  # type: ignore[arg-type]
            origin = get_origin(self.type_hint)
            type_hint_name = self.type_hint.__name__ if origin is None else origin.__name__ if hasattr(origin, '__name__') else repr(origin)  # type: ignore[attr-defined]
            self.raise_exception_in_storage(TypeError(f'The value {self.get_value_representation(value)} of the {self.get_field_name_representation()} does not match the type {type_hint_name}.'), raise_all)

    def get_field_name_representation(self) -> str:
        if self.doc is None:
            return f'"{self.name}" field'
        return f'"{self.name}" field ({self.doc})'

    def check_value(self, value: ValueType, raise_all: bool = False) -> None:
        if self.validation is not None:
            if isinstance(self.validation, dict):
                for message, validator in self.validation.items():
                    if not validator(value):
                        self.raise_exception_in_storage(ValueError(message), raise_all)
            elif not self.validation(value):
                self.raise_exception_in_storage(ValueError(f'The value {self.get_value_representation(value)} of the {self.get_field_name_representation()} does not match the validation.'), raise_all)

    def get_field_lock(self, instance: Storage) -> ContextLockProtocol:
        return instance.__locks__[cast(str, self.name)]

    def get_value_representation(self, value: ValueType) -> str:
        base = '***' if self.secret else f'{value!r}'
        return f'{base} ({type(value).__name__})'

    def raise_exception_in_storage(self, exception: BaseException, raising_on: bool) -> None:
        if raising_on:
            raise exception
        self.exception = exception

    def get_sources(self, instance: Storage) -> SourcesCollection[ExpectedType]:
        if self.sources is None:
            return instance.__sources__

        result = []
        there_is_ellipsis = False

        for source in self.sources:
            if source is Ellipsis:
                there_is_ellipsis = True
            else:
                result.append(source)

        if there_is_ellipsis:
           result.extend(instance.__sources__.sources)

        return cast(SourcesCollection[ExpectedType], SourcesCollection(result))



from abc import ABC, abstractmethod
from typing import Generic, Optional, Type, TypeVar, cast

from denial import InnerNoneType
from simtypes import check

ExpectedType = TypeVar('ExpectedType')
sentinel = InnerNoneType()

class AbstractSource(Generic[ExpectedType], ABC):
    @abstractmethod
    def __getitem__(self, key: str) -> ExpectedType:
        ...  # pragma: no cover

    def get(self, key: str, default: Optional[ExpectedType] = None) -> Optional[ExpectedType]:
        try:
            result: ExpectedType = self[key]
        except KeyError:
            return default

        return result

    def type_awared_get(self, key: str, hint: Type[ExpectedType], default: ExpectedType = cast(ExpectedType, sentinel)) -> Optional[ExpectedType]:  # noqa: B008
        result = self.get(key, default)

        if result is default:
            if default is sentinel:
                return None
            return default

        if not check(result, hint, strict=True):
            raise TypeError(f'The value of the "{key}" field did not pass the type check.')

        return result



import sys
from typing import Any, List, Optional, Union

import pytest
from full_match import match
from locklib import LockTraceWrapper
from sigmatch import SignatureMismatchError

from skelet import (
    EnvSource,
    Field,
    JSONSource,
    MemorySource,
    NaturalNumber,
    NonNegativeInt,
    Storage,
    TOMLSource,
    YAMLSource,
)


def test_try_to_get_descriptor_object_from_class_inherited_from_storage():
    class SomeClass(Storage):
        field = Field(42)

    assert isinstance(SomeClass.field, Field)


def test_try_to_use_field_outside_storage():
    if sys.version_info < (3, 12):
        with pytest.raises(RuntimeError):
            class SomeClass:
                field = Field(42)
    else:
        with pytest.raises(TypeError):
            class SomeClass:
                field = Field(42)


def test_try_to_use_one_field_in_two_storage_classes():
    class FirstClass(Storage):
        field = Field(42)

    with pytest.raises(TypeError):
        class SecondClass(Storage):
            field = FirstClass.__dict__['field']


def test_set_default_value_and_read_it():
    class SomeClass(Storage):
        field = Field(42)

    some_object = SomeClass()

    assert some_object.field == 42


def test_set_not_default_value_and_read_it():
    class SomeClass(Storage):
        field = Field(42)

    object_1 = SomeClass()
    object_2 = SomeClass()

    assert object_1.field == 42
    assert object_2.field == 42

    object_1.field = 100
    object_2.field = 200

    assert object_1.field == 100
    assert object_2.field == 200


def test_set_not_default_value_and_read_multiple_times():
    class SomeClass(Storage):
        field = Field(0)

    instance = SomeClass()

    for index in range(10):
        assert instance.field == index
        instance.field += 1


def test_changing_value_is_not_changing_the_default_value():
    class SomeClass(Storage):
        field = Field(42)

    instance = SomeClass()

    assert instance.field == 42

    instance.field += 1

    assert instance.field == 43

    assert SomeClass().field == 42


def test_try_to_delete_field():
    class SomeClass(Storage):
        field = Field(42)

    with pytest.raises(AttributeError, match=match('You can\'t delete the "field" field value.')):
        del SomeClass().field


def test_try_to_delete_field_with_doc():
    class SomeClass(Storage):
        field = Field(42, doc='some doc')

    with pytest.raises(AttributeError, match=match('You can\'t delete the "field" field (some doc) value.')):
        del SomeClass().field


def test_try_to_set_new_value_to_read_only_attribute():
    class SomeClass(Storage):
        field = Field(42, read_only=True)

    instance = SomeClass()

    with pytest.raises(AttributeError, match=match('"field" field is read-only.')):
        instance.field = 43

    assert instance.field == 42


def test_try_to_set_new_value_to_read_only_attribute_with_doc():
    class SomeClass(Storage):
        field = Field(42, read_only=True, doc='some doc')

    instance = SomeClass()

    with pytest.raises(AttributeError, match=match('"field" field (some doc) is read-only.')):
        instance.field = 43

    assert instance.field == 42


def test_all_storage_childs_have_their_own_lists_with_names():
    class FirstClass(Storage):
        field_1 = Field(42)
        field_2 = Field(43)
        field_3 = Field(44)

    class SecondClass(Storage):
        field_1 = Field(42)
        field_2 = Field(43)
        field_3 = Field(44)

    assert FirstClass.__field_names__ == ['field_1', 'field_2', 'field_3']
    assert SecondClass.__field_names__ == ['field_1', 'field_2', 'field_3']

    assert FirstClass.__field_names__ is not SecondClass.__field_names__

    assert FirstClass().field_1 == 42
    assert FirstClass().field_2 == 43
    assert FirstClass().field_3 == 44
    assert SecondClass().field_1 == 42
    assert SecondClass().field_2 == 43
    assert SecondClass().field_3 == 44


def test_inheritance_of_fields():
    class FirstClass(Storage):
        field_1 = Field(42)
        field_2 = Field(43)
        field_3 = Field(44)

    class SecondClass(FirstClass):
        ...

    assert FirstClass.__field_names__ == ['field_1', 'field_2', 'field_3']
    assert SecondClass.__field_names__ == FirstClass.__field_names__

    assert FirstClass().field_1 == 42
    assert FirstClass().field_2 == 43
    assert FirstClass().field_3 == 44
    assert SecondClass().field_1 == 42
    assert SecondClass().field_2 == 43
    assert SecondClass().field_3 == 44


def test_inheritance_of_fields_and_adding_new_fields():
    class FirstClass(Storage):
        field_1 = Field(42)
        field_2 = Field(43)
        field_3 = Field(44)

    class SecondClass(FirstClass):
        field_4 = Field(45)

    assert FirstClass.__field_names__ == ['field_1', 'field_2', 'field_3']
    assert SecondClass.__field_names__ ==  [*FirstClass.__field_names__, 'field_4']

    assert FirstClass().field_1 == 42
    assert FirstClass().field_2 == 43
    assert FirstClass().field_3 == 44
    assert SecondClass().field_1 == 42
    assert SecondClass().field_2 == 43
    assert SecondClass().field_3 == 44
    assert SecondClass().field_4 == 45


def test_inheritance_of_fields_and_adding_new_fields_two_times():
    class FirstClass(Storage):
        field_1 = Field(42)
        field_2 = Field(43)
        field_3 = Field(44)

    class SecondClass(FirstClass):
        field_4 = Field(45)

    class ThirdClass(SecondClass):
        field_5 = Field(46)

    assert FirstClass.__field_names__ == ['field_1', 'field_2', 'field_3']
    assert SecondClass.__field_names__ ==  [*FirstClass.__field_names__, 'field_4']
    assert ThirdClass.__field_names__ == [*SecondClass.__field_names__, 'field_5']

    assert FirstClass().field_1 == 42
    assert FirstClass().field_2 == 43
    assert FirstClass().field_3 == 44
    assert SecondClass().field_1 == 42
    assert SecondClass().field_2 == 43
    assert SecondClass().field_3 == 44
    assert SecondClass().field_4 == 45
    assert ThirdClass().field_1 == 42
    assert ThirdClass().field_2 == 43
    assert ThirdClass().field_3 == 44
    assert ThirdClass().field_4 == 45
    assert ThirdClass().field_5 == 46


def test_redefine_field_in_child_class():
    class FirstClass(Storage):
        field = Field(42)

    class SecondClass(Storage):
        field = Field(43)

    assert FirstClass.__field_names__ == ['field']
    assert SecondClass.__field_names__ == ['field']

    assert FirstClass().field == 42
    assert SecondClass().field == 43


def test_redefine_field_in_child_class_and_change_value():
    class FirstClass(Storage):
        field = Field(42)

    class SecondClass(Storage):
        field = Field(43)

    assert FirstClass.__field_names__ == ['field']
    assert SecondClass.__field_names__ == ['field']

    first = FirstClass()
    second = SecondClass()

    assert first.field == 42
    assert second.field == 43

    first.field = 44

    assert first.field == 44
    assert second.field == 43

    second.field = 45

    assert first.field == 44
    assert second.field == 45


def test_storage_child_has_fields_list():
    class StorageChild(Storage):
        ...

    assert StorageChild.__field_names__ == ()


def test_repr_without_fields():
    class StorageChild(Storage):
        ...

    assert repr(StorageChild()) == 'StorageChild()'


def test_repr_with_fields():
    class StorageChild(Storage):
        field_1 = Field(42)
        field_2 = Field(43)

    assert repr(StorageChild()) == 'StorageChild(field_1=42, field_2=43)'


def test_repr_with_fields_and_values():
    class StorageChild(Storage):
        field_1 = Field(42)
        field_2 = Field(43)

    assert repr(StorageChild()) == 'StorageChild(field_1=42, field_2=43)'
    assert repr(StorageChild(field_1=44, field_2=45)) == 'StorageChild(field_1=44, field_2=45)'


def test_set_some_values_in_init():
    class StorageChild(Storage):
        field_1 = Field(42)
        field_2 = Field(43)

    storage = StorageChild(field_1=44)

    assert storage.field_1 == 44
    assert storage.field_2 == 43

    assert repr(storage) == 'StorageChild(field_1=44, field_2=43)'


def test_try_to_set_not_defined_field_in_init():
    class StorageChild(Storage):
        field_1 = Field(42)
        field_2 = Field(43)

    with pytest.raises(KeyError, match=r'The "field_3" field is not defined.'):
        StorageChild(field_3=44)


def test_get_from_inner_dict_is_thread_safe_and_use_per_fields_locks():
    class SomeClass(Storage):
        field = Field(42, read_lock=True)

    storage = SomeClass()
    field = SomeClass.field

    field.lock = LockTraceWrapper(field.lock)
    storage.__locks__['field'] = LockTraceWrapper(storage.__locks__['field'])

    class PseudoDict:
        def get(self, key):  # noqa: ARG002
            storage.__locks__['field'].notify('get')
            field.lock.notify('get')
            return 43

    storage.__values__ = PseudoDict()

    assert storage.field == 43
    assert storage.__locks__['field'].was_event_locked('get')

    assert not field.lock.was_event_locked('get')
    assert field.lock.trace


def test_that_set_is_thread_safe_and_use_per_field_locks():
    class SomeClass(Storage):
        field = Field(42)

    storage = SomeClass()
    field = SomeClass.field

    field.lock = LockTraceWrapper(field.lock)
    storage.__locks__['field'] = LockTraceWrapper(storage.__locks__['field'])
    class PseudoDict:
        def __setitem__(self, key, default):
            storage.__locks__['field'].notify('set')
            field.lock.notify('set')

        def get(self, key):  # noqa: ARG002
            storage.__locks__['field'].notify('get')
            field.lock.notify('get')
            return 42

    storage.__values__ = PseudoDict()

    storage.field = 44

    assert storage.__locks__['field'].was_event_locked('set')
    assert storage.__locks__['field'].was_event_locked('get')

    assert not field.lock.was_event_locked('set')
    assert field.lock.trace
    assert not field.lock.was_event_locked('get')


def test_set_name_uses_per_field_object_lock():
    class SomeClass(Storage):
        ...

    field = Field(42)
    field.lock = LockTraceWrapper(field.lock)
    field.set_field_names = lambda x, y: field.lock.notify('get')  # noqa: ARG005

    field.__set_name__(SomeClass, 'field')

    assert field.lock.was_event_locked('get')
    assert field.lock.trace


def test_simple_type_check_failed_when_set_bool_if_expected_int():
    class SomeClass(Storage):
        field: int = Field(15)

    instance = SomeClass()

    instance.field = True

    assert instance.field is True


@pytest.mark.parametrize(
    ('int_value', 'float_value', 'secret'),
    [
        ('***', '***', True),
        ("'15'", '15.0', False),
    ],
)
def test_simple_type_check_failed_when_set(int_value, float_value, secret):
    class SomeClass(Storage):
        field: int = Field(15, secret=secret)

    instance = SomeClass()

    with pytest.raises(TypeError, match=match(f'The value {int_value} (str) of the "field" field does not match the type int.')):
        instance.field = '15'

    with pytest.raises(TypeError, match=match(f'The value {float_value} (float) of the "field" field does not match the type int.')):
        instance.field = 15.0

    assert instance.field == 15
    assert type(instance.field) is int


@pytest.mark.parametrize(
    ('int_value', 'float_value', 'secret'),
    [
        ('***', '***', True),
        ("'15'", "15.0", False),
    ],
)
def test_simple_type_check_failed_when_set_with_doc(int_value, float_value, secret):
    class SomeClass(Storage):
        field: int = Field(15, doc='some doc', secret=secret)

    instance = SomeClass()

    with pytest.raises(TypeError, match=match(f'The value {int_value} (str) of the "field" field (some doc) does not match the type int.')):
        instance.field = '15'

    with pytest.raises(TypeError, match=match(f'The value {float_value} (float) of the "field" field (some doc) does not match the type int.')):
        instance.field = 15.0

    assert instance.field == 15
    assert type(instance.field) is int


def test_simple_type_check_not_failed_when_set():
    class SomeClass(Storage):
        field: int = Field(15)

    instance = SomeClass()

    instance.field = 16

    assert instance.field == 16
    assert type(instance.field) is int


@pytest.mark.parametrize(
    ('wrong_value', 'secret'),
    [
        ('***', True),
        ("'15'", False),
    ],
)
def test_type_check_when_define_default_failed(wrong_value, secret):
    with pytest.raises(TypeError, match=match(f'The value {wrong_value} (str) of the "field" field does not match the type int.')):
        class SomeClass(Storage):
            field: int = Field('15', secret=secret)


@pytest.mark.parametrize(
    ('wrong_value', 'secret'),
    [
        ('***', True),
        ("'15'", False),
    ],
)
def test_type_check_when_define_default_failed_with_doc(wrong_value, secret):
    with pytest.raises(TypeError, match=match(f'The value {wrong_value} (str) of the "field" field (some doc) does not match the type int.')):
        class SomeClass(Storage):
            field: int = Field('15', doc='some doc', secret=secret)


def test_type_check_when_define_default_not_failed():
    class SomeClass(Storage):
        field: int = Field(15)

    assert SomeClass().field == 15
    assert type(SomeClass().field) is int


@pytest.mark.parametrize(
    ('wrong_value', 'secret'),
    [
        ('***', True),
        ("'kek'", False),
    ],
)
def test_type_check_when_redefine_defaults_initing_new_object_failed(wrong_value, secret):
    class SomeClass(Storage):
        field: int = Field(15, secret=secret)

    with pytest.raises(TypeError, match=match(f'The value {wrong_value} (str) of the "field" field does not match the type int.')):
        SomeClass(field='kek')


@pytest.mark.parametrize(
    ('wrong_value', 'secret'),
    [
        ('***', True),
        ("'kek'", False),
    ],
)
def test_type_check_when_redefine_defaults_initing_new_object_failed_with_doc(wrong_value, secret):
    class SomeClass(Storage):
        field: int = Field(15, doc='some doc', secret=secret)

    with pytest.raises(TypeError, match=match(f'The value {wrong_value} (str) of the "field" field (some doc) does not match the type int.')):
        SomeClass(field='kek')


def test_type_check_when_redefine_defaults_initing_new_object_not_failed():
    class SomeClass(Storage):
        field: int = Field(15)

    instance = SomeClass(field=16)

    assert instance.field == 16
    assert type(instance.field) is int

    instance = SomeClass(field=-100)

    assert instance.field == -100
    assert type(instance.field) is int


@pytest.mark.parametrize(
    ('wrong_value', 'secret'),
    [
        ('***', True),
        ("'kek'", False),
    ],
)
def test_more_examples_of_type_check_when_redefine_defaults_initing_new_object_failed(wrong_value, secret):
    class SomeClass(Storage):
        field: Optional[int] = Field(15, secret=secret)

    if sys.version_info < (3, 10):
        type_representation = 'typing.Union'
    else:
        type_representation = 'Union'

    with pytest.raises(TypeError, match=match(f'The value {wrong_value} (str) of the "field" field does not match the type {type_representation}.')):
        SomeClass(field='kek')

    instance = SomeClass(field=None)

    assert instance.field is None

    instance = SomeClass(field=1000)

    assert instance.field == 1000

    class SecondClass(Storage):
        field: Any = Field(15)

    instance = SecondClass(field='kek')

    assert instance.field == 'kek'

    instance = SecondClass(field=None)

    assert instance.field is None

    instance = SecondClass(field=1000)

    assert instance.field == 1000


@pytest.mark.parametrize(
    ('wrong_value', 'secret'),
    [
        ('***', True),
        ("'kek'", False),
    ],
)
def test_more_examples_of_type_check_when_redefine_defaults_initing_new_object_failed_with_doc(wrong_value, secret):
    class SomeClass(Storage):
        field: Optional[int] = Field(15, doc='some doc', secret=secret)

    if sys.version_info < (3, 10):
        type_representation = 'typing.Union'
    else:
        type_representation = 'Union'

    with pytest.raises(TypeError, match=match(f'The value {wrong_value} (str) of the "field" field (some doc) does not match the type {type_representation}.')):
        SomeClass(field='kek')


def test_try_to_use_underscored_name_for_field():
    with pytest.raises(ValueError, match=match('Field name "_field" cannot start with an underscore.')):
        class SomeClass(Storage):
            _field: int = Field(15)


def test_try_to_use_underscored_name_for_field_with_doc():
    with pytest.raises(ValueError, match=match('Field name "_field" cannot start with an underscore.')):
        class SomeClass(Storage):
            _field: int = Field(15, doc='some doc')


@pytest.mark.parametrize(
    ('wrong_value', 'secret'),
    [
        ('***', True),
        ('-1', False),
    ],
)
def test_validation_function_failed_when_set(wrong_value, secret):
    class SomeClass(Storage):
        field: int = Field(15, validation=lambda value: value > 0, secret=secret)

    instance = SomeClass()

    with pytest.raises(ValueError, match=match(f'The value {wrong_value} (int) of the "field" field does not match the validation.')):
        instance.field = -1


@pytest.mark.parametrize(
    ('wrong_value', 'secret'),
    [
        ('***', True),
        ('-1', False),
    ],
)
def test_validation_function_failed_when_set_with_doc(wrong_value, secret):
    class SomeClass(Storage):
        field: int = Field(15, validation=lambda value: value > 0, doc='some doc', secret=secret)

    instance = SomeClass()

    with pytest.raises(ValueError, match=match(f'The value {wrong_value} (int) of the "field" field (some doc) does not match the validation.')):
        instance.field = -1


@pytest.mark.parametrize(
    'addictional_parameters',
    [
        {},
        {'doc': 'some doc'},
    ],
)
def test_validation_functions_dict_failed_when_set(addictional_parameters):
    class SomeClass(Storage):
        field: int = Field(15, validation={'some message': lambda x: x > 0}, **addictional_parameters)

    instance = SomeClass()

    with pytest.raises(ValueError, match=match('some message')):
        instance.field = -1


def test_validation_function_not_failed_when_set():
    class SomeClass(Storage):
        field: int = Field(15, validation=lambda value: value > 0)

    instance = SomeClass()

    instance.field = 1

    assert instance.field == 1


def test_validation_functions_dict_not_failed_when_set():
    class SomeClass(Storage):
        field: int = Field(15, validation={'some message': lambda value: value > 0})

    instance = SomeClass()

    instance.field = 1

    assert instance.field == 1


@pytest.mark.parametrize(
    ('wrong_value', 'secret'),
    [
        ('***', True),
        ('-1', False),
    ],
)
def test_validation_function_failed_when_init(wrong_value, secret):
    class SomeClass(Storage):
        field: int = Field(15, validation=lambda value: value > 0, secret=secret)

    with pytest.raises(ValueError, match=match(f'The value {wrong_value} (int) of the "field" field does not match the validation.')):
        SomeClass(field=-1)


@pytest.mark.parametrize(
    ('wrong_value', 'secret'),
    [
        ('***', True),
        ('-1', False),
    ],
)
def test_validation_function_failed_when_init_with_doc(wrong_value, secret):
    class SomeClass(Storage):
        field: int = Field(15, validation=lambda value: value > 0, doc='some doc', secret=secret)

    with pytest.raises(ValueError, match=match(f'The value {wrong_value} (int) of the "field" field (some doc) does not match the validation.')):
        SomeClass(field=-1)


@pytest.mark.parametrize(
    'addictional_parameters',
    [
        {},
        {'doc': 'some doc'},
    ],
)
def test_validation_functions_dict_failed_when_init(addictional_parameters):
    class SomeClass(Storage):
        field: int = Field(15, validation={'some message': lambda value: value > 0}, **addictional_parameters)

    with pytest.raises(ValueError, match=match('some message')):
        SomeClass(field=-1)


@pytest.mark.parametrize(
    'addictional_parameters',
    [
        {},
        {'doc': 'some doc'},
    ],
)
def test_validation_function_not_failed_when_init(addictional_parameters):
    class SomeClass(Storage):
        field: int = Field(15, validation=lambda value: value > 0, **addictional_parameters)

    instance = SomeClass()

    instance.field = 1

    assert instance.field == 1


@pytest.mark.parametrize(
    'addictional_parameters',
    [
        {},
        {'doc': 'some doc'},
    ],
)
def test_validation_functions_dict_not_failed_when_init(addictional_parameters):
    class SomeClass(Storage):
        field: int = Field(15, validation={'some message': lambda value: value > 0}, **addictional_parameters)

    instance = SomeClass()

    instance.field = 1

    assert instance.field == 1


@pytest.mark.parametrize(
    ('wrong_value', 'secret'),
    [
        ('***', True),
        ('-15', False),
    ],
)
def test_validation_function_failed_when_default(wrong_value, secret):
    with pytest.raises(ValueError, match=match(f'The value {wrong_value} (int) of the "field" field does not match the validation.')):
        class SomeClass(Storage):
            field: int = Field(-15, validation=lambda value: value > 0, secret=secret)


@pytest.mark.parametrize(
    ('wrong_value', 'secret'),
    [
        ('***', True),
        ('-15', False),
    ],
)
def test_validation_function_failed_when_default_with_doc(wrong_value, secret):
    with pytest.raises(ValueError, match=match(f'The value {wrong_value} (int) of the "field" field (some doc) does not match the validation.')):
        class SomeClass(Storage):
            field: int = Field(-15, validation=lambda value: value > 0, doc='some doc', secret=secret)


@pytest.mark.parametrize(
    'addictional_parameters',
    [
        {},
        {'doc': 'some doc'},
    ],
)
def test_validation_function_not_failed_when_default_because_no_check_first_time(addictional_parameters):
    class SomeClass(Storage):
        field: int = Field(-15, validation=lambda value: value > 0, validate_default=False, **addictional_parameters)

    assert SomeClass().field == -15


def test_validation_when_set_is_not_under_lock():
    class SomeClass(Storage):
        field: int = Field(10, validation=lambda value: value > 0)

    instance = SomeClass()

    instance.__locks__['field'] = LockTraceWrapper(instance.__locks__['field'])
    SomeClass.field.validation = lambda x: instance.__locks__['field'].notify('kek') is None  # noqa: ARG005
    instance.field = 5

    assert instance.field == 5

    assert not instance.__locks__['field'].was_event_locked('kek')


def test_type_check_when_set_is_not_under_lock():
    class SomeClass(Storage):
        field: int = Field(10, validation=lambda value: value > 0)

    instance = SomeClass()

    instance.__locks__['field'] = LockTraceWrapper(instance.__locks__['field'])
    SomeClass.field.check_type_hints = lambda x, raise_all: instance.__locks__['field'].notify('kek')  # noqa: ARG005
    instance.field = 5

    assert instance.field == 5

    assert not instance.__locks__['field'].was_event_locked('kek')


def test_type_check_when_set_is_before_validation():
    flags = []
    start_check = False

    def validation(value):
        nonlocal flags
        if start_check:
            flags.append('validation')

        return isinstance(value, int)

    class SomeClass(Storage):
        field: int = Field(10, validation=validation)

    instance = SomeClass()

    old_check_type_hints = SomeClass.field.check_type_hints
    SomeClass.field.check_type_hints = lambda z, raise_all: flags.append('type_check') is old_check_type_hints(z, raise_all=raise_all)
    start_check = True

    with pytest.raises(TypeError):
        instance.field = 'kek'

    assert instance.field == 10
    assert flags == ['type_check']

    SomeClass.field.check_type_hints = old_check_type_hints


def test_repr_for_secret_fields():
    class SomeClass(Storage):
        field: int = Field(10, secret=True)
        second_field: int = Field(100)

    instance = SomeClass()

    assert repr(instance) == 'SomeClass(field=***, second_field=100)'

    instance.field = instance.field * 2
    instance.second_field = instance.second_field * 2

    assert repr(instance) == 'SomeClass(field=***, second_field=200)'


def test_change_value_of_secret_field():
    class SomeClass(Storage):
        field: int = Field(10, secret=True)

    instance = SomeClass()

    assert instance.field == 10

    instance.field = 20

    assert instance.field == 20


def test_change_value_of_secret_field_in_init():
    class SomeClass(Storage):
        field: int = Field(10, secret=True)

    instance = SomeClass(field=20)

    assert instance.field == 20


def test_set_action_for_set():
    flags = []

    class SomeClass(Storage):
        field: int = Field(10, secret=True, action=lambda old, new, storage: flags.append(True))  # noqa: ARG005

    instance = SomeClass()

    assert not flags

    instance.field = 13

    assert flags == [True]

    instance.field = 14

    assert flags == [True, True]


def test_action_doesnt_work_when_new_value_is_same():
    flags = []

    class SomeClass(Storage):
        field: int = Field(10, secret=True, action=lambda old, new, storage: flags.append(True))  # noqa: ARG005

    instance = SomeClass()

    assert not flags

    instance.field = 10

    assert not flags
    assert instance.field == 10


@pytest.mark.parametrize(
    'addictional_arguments',
    [
        {'read_lock': True},
    ],
)
def test_read_lock_on(addictional_arguments):
    class SomeClass(Storage):
        field: int = Field(10, secret=True, **addictional_arguments)

    instance = SomeClass()

    lock = LockTraceWrapper(instance.__locks__['field'])
    instance.__locks__['field'] = lock
    field = SomeClass.field
    field.lock = LockTraceWrapper(field.lock)

    class PseudoDict:
        def get(self, key):  # noqa: ARG002
            lock.notify('get')
            field.lock.notify('get')
            return 10

    instance.__values__ = PseudoDict()

    assert not lock.trace
    assert not field.lock.trace

    assert instance.field == 10

    assert lock.trace
    assert field.lock.trace

    assert lock.was_event_locked('get')
    assert not field.lock.was_event_locked('get')


def test_read_lock_off():
    class SomeClass(Storage):
        field: int = Field(10, secret=True, read_lock=False)

    instance = SomeClass()

    lock = LockTraceWrapper(instance.__locks__['field'])
    instance.__locks__['field'] = lock
    field = SomeClass.field
    field.lock = LockTraceWrapper(field.lock)

    class PseudoDict:
        def get(self, key):  # noqa: ARG002
            lock.notify('get')
            field.lock.notify('get')
            return 10

    instance.__values__ = PseudoDict()

    assert not lock.trace
    assert not field.lock.trace

    assert instance.field == 10

    assert lock.trace
    assert field.lock.trace

    assert not lock.was_event_locked('get')
    assert not field.lock.was_event_locked('get')


def test_two_storage_instances_by_default_have_not_the_same_locks():
    class SomeClass(Storage):
        field: int = Field(10)
        other_field: int = Field(20)

    instance = SomeClass()
    second_instance = SomeClass()

    assert instance.__locks__ is not second_instance.__locks__

    assert instance.__locks__['field'] is not instance.__locks__['other_field']
    assert second_instance.__locks__['field'] is not second_instance.__locks__['other_field']

    assert second_instance.__locks__['field'] is not instance.__locks__['field']
    assert second_instance.__locks__['other_field'] is not instance.__locks__['other_field']


def test_storage_is_not_singleton():
    class SomeClass(Storage):
        field: int = Field(10)

    instance = SomeClass()
    second_instance = SomeClass()

    assert instance is not second_instance


def test_conflicting_fields_have_the_same_lock():
    class SomeClass(Storage):
        field: int = Field(10, conflicts={'other_field': lambda old, new, other_old, other_new: new > other_old})  # noqa: ARG005
        other_field: int = Field(20)
        second_other_field: int = Field(25)

    instance = SomeClass()

    assert instance.__locks__['field'] is instance.__locks__['other_field']
    assert instance.__locks__['field'] is not instance.__locks__['second_other_field']


def test_conflicts_check_is_under_field_lock():
    locks: List[LockTraceWrapper] = []

    def check_function(old, new, other_old, other_new):  # noqa: ARG001
        for lock in locks:
            lock.notify('check')
        return False

    class SomeClass(Storage):
        field: int = Field(10, conflicts={'other_field': check_function})
        other_field: int = Field(20)

    instance = SomeClass()

    lock = LockTraceWrapper(instance.__locks__['field'])
    locks.append(lock)
    instance.__locks__['field'] = lock

    instance.field = 20

    assert lock.trace
    assert lock.was_event_locked('check')


def test_reverse_conflicts_check_is_under_field_lock():
    locks: List[LockTraceWrapper] = []

    def check_function(old, new, other_old, other_new):  # noqa: ARG001
        for lock in locks:
            lock.notify('check')
        return False

    class SomeClass(Storage):
        field: int = Field(10, conflicts={'other_field': check_function})
        other_field: int = Field(20)

    instance = SomeClass()

    assert instance.__locks__['field'] is instance.__locks__['other_field']

    lock = LockTraceWrapper(instance.__locks__['other_field'])
    locks.append(lock)
    instance.__locks__['other_field'] = lock

    instance.other_field = 25

    assert lock.trace
    assert lock.was_event_locked('check')


@pytest.mark.parametrize(
    'addictional_arguments',
    [
        {},
        {'doc': 'some doc'},
    ],
)
def test_non_existing_conflicting_field_name(addictional_arguments):
    if not addictional_arguments:
        exception_message = 'You have set a conflict condition for "field" field with field "ather_field", but the field "ather_field" does not exist in the class SomeClass.'
    else:
        exception_message = f'You have set a conflict condition for "field" field ({addictional_arguments["doc"]}) with field "ather_field", but the field "ather_field" does not exist in the class SomeClass.'

    with pytest.raises(NameError, match=match(exception_message)):
        class SomeClass(Storage):
            field: int = Field(10, conflicts={'ather_field': lambda old, new, other_old, other_new: new > other_old}, **addictional_arguments)  # noqa: ARG005
            other_field: int = Field(20)


# Check: reverse check
# Check: exceptions messages for both types of fields on the both sides, direct and reverse

@pytest.mark.parametrize(
    'main_field_is_secret',
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    'addictional_arguments',
    [
        {},
        {'doc': 'some doc'},
    ],
)
def test_basic_conflicting_fields(addictional_arguments, main_field_is_secret):
    class SomeClass(Storage):
        field: int = Field(10, conflicts={'other_field': lambda old, new, other_old, other_new: new > other_old, 'secret_other_field': lambda old, new, other_old, other_new: new < 0}, doc=addictional_arguments.get('doc'), secret=main_field_is_secret)  # noqa: ARG005
        other_field: int = Field(20, doc=addictional_arguments.get('doc'))
        secret_other_field: int = Field(20, secret=True, doc=addictional_arguments.get('doc'))

    instance = SomeClass()

    assert instance.field == 10

    instance.field = 15

    assert instance.field == 15

    if 'doc' in addictional_arguments:
        if main_field_is_secret:
            exception_message = 'The new *** (int) value of the "field" field (some doc) conflicts with the 20 (int) value of the "other_field" field (some doc).'
        else:
            exception_message = 'The new 21 (int) value of the "field" field (some doc) conflicts with the 20 (int) value of the "other_field" field (some doc).'
    elif main_field_is_secret:
        exception_message = 'The new *** (int) value of the "field" field conflicts with the 20 (int) value of the "other_field" field.'
    else:
        exception_message = 'The new 21 (int) value of the "field" field conflicts with the 20 (int) value of the "other_field" field.'

    with pytest.raises(ValueError, match=match(exception_message)):
        instance.field = 21

    assert instance.field == 15

    if 'doc' in addictional_arguments:
        if main_field_is_secret:
            exception_message = 'The new *** (int) value of the "field" field (some doc) conflicts with the *** (int) value of the "secret_other_field" field (some doc).'
        else:
            exception_message = 'The new -1 (int) value of the "field" field (some doc) conflicts with the *** (int) value of the "secret_other_field" field (some doc).'
    elif main_field_is_secret:
        exception_message = 'The new *** (int) value of the "field" field conflicts with the *** (int) value of the "secret_other_field" field.'
    else:
        exception_message = 'The new -1 (int) value of the "field" field conflicts with the *** (int) value of the "secret_other_field" field.'

    with pytest.raises(ValueError, match=match(exception_message)):
        instance.field = -1

    assert instance.field == 15


@pytest.mark.parametrize(
    'main_field_is_secret',
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    'addictional_arguments',
    [
        {},
        {'doc': 'some doc'},
    ],
)
def test_conflicting_fields_when_set_in_init(addictional_arguments, main_field_is_secret):
    class SomeClass(Storage):
        field: int = Field(10, conflicts={'other_field': lambda old, new, other_old, other_new: new > other_old, 'secret_other_field': lambda old, new, other_old, other_new: new < 0}, doc=addictional_arguments.get('doc'), secret=main_field_is_secret)  # noqa: ARG005
        other_field: int = Field(20, doc=addictional_arguments.get('doc'))
        secret_other_field: int = Field(20, secret=True, doc=addictional_arguments.get('doc'))

    instance = SomeClass()

    assert instance.field == 10

    instance = SomeClass(field=15)

    assert instance.field == 15

    if 'doc' in addictional_arguments:
        if main_field_is_secret:
            exception_message = 'The new *** (int) value of the "field" field (some doc) conflicts with the 20 (int) value of the "other_field" field (some doc).'
        else:
            exception_message = 'The new 21 (int) value of the "field" field (some doc) conflicts with the 20 (int) value of the "other_field" field (some doc).'
    elif main_field_is_secret:
        exception_message = 'The new *** (int) value of the "field" field conflicts with the 20 (int) value of the "other_field" field.'
    else:
        exception_message = 'The new 21 (int) value of the "field" field conflicts with the 20 (int) value of the "other_field" field.'

    with pytest.raises(ValueError, match=match(exception_message)):
        SomeClass(field=21)

    if 'doc' in addictional_arguments:
        if main_field_is_secret:
            exception_message = 'The new *** (int) value of the "field" field (some doc) conflicts with the *** (int) value of the "secret_other_field" field (some doc).'
        else:
            exception_message = 'The new -1 (int) value of the "field" field (some doc) conflicts with the *** (int) value of the "secret_other_field" field (some doc).'
    elif main_field_is_secret:
        exception_message = 'The new *** (int) value of the "field" field conflicts with the *** (int) value of the "secret_other_field" field.'
    else:
        exception_message = 'The new -1 (int) value of the "field" field conflicts with the *** (int) value of the "secret_other_field" field.'

    with pytest.raises(ValueError, match=match(exception_message)):
        SomeClass(field=-1)


@pytest.mark.parametrize(
    'are_fields_secret',
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    'addictional_arguments',
    [
        {},
        {'doc': 'some doc'},
    ],
)
def test_conflicting_fields_when_defaults_are_conflicting(addictional_arguments, are_fields_secret):
    if 'doc' in addictional_arguments:
        if are_fields_secret:
            exception_message = 'The *** (int) default value of the "field" field (some doc) conflicts with the *** (int) value of the "other_field" field (some doc).'
        else:
            exception_message = 'The 21 (int) default value of the "field" field (some doc) conflicts with the 20 (int) value of the "other_field" field (some doc).'
    elif are_fields_secret:
        exception_message = 'The *** (int) default value of the "field" field conflicts with the *** (int) value of the "other_field" field.'
    else:
        exception_message = 'The 21 (int) default value of the "field" field conflicts with the 20 (int) value of the "other_field" field.'

    with pytest.raises(ValueError, match=match(exception_message)):
        class SomeClass(Storage):
            field: int = Field(21, conflicts={'other_field': lambda old, new, other_old, other_new: new > other_old, 'secret_other_field': lambda old, new, other_old, other_new: new > 30}, doc=addictional_arguments.get('doc'), secret=are_fields_secret)  # noqa: ARG005
            other_field: int = Field(20, doc=addictional_arguments.get('doc'), secret=are_fields_secret)


@pytest.mark.parametrize(
    'are_fields_secret',
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    'addictional_arguments',
    [
        {},
        {'doc': 'some doc'},
        {'reverse_conflicts': True},
        {'reverse_conflicts': True, 'doc': 'some doc'},
    ],
)
def test_basic_conflicting_fields_reverse_when_its_on(addictional_arguments, are_fields_secret):
    doc = addictional_arguments.pop('doc', None)

    class SomeClass(Storage):
        field: int = Field(10, conflicts={'other_field': lambda old, new, other_old, other_new: new > other_new}, doc=doc, secret=are_fields_secret, **addictional_arguments)  # noqa: ARG005
        other_field: int = Field(20, doc=doc, secret=are_fields_secret, **addictional_arguments)

    instance = SomeClass()

    assert instance.field == 10
    assert instance.other_field == 20

    instance.other_field = 30

    assert instance.other_field == 30

    if doc is not None:
        if are_fields_secret:
            exception_message = 'The new *** (int) value of the "other_field" field (some doc) conflicts with the *** (int) value of the "field" field (some doc).'
        else:
            exception_message = 'The new 5 (int) value of the "other_field" field (some doc) conflicts with the 10 (int) value of the "field" field (some doc).'
    elif are_fields_secret:
        exception_message = 'The new *** (int) value of the "other_field" field conflicts with the *** (int) value of the "field" field.'
    else:
        exception_message = 'The new 5 (int) value of the "other_field" field conflicts with the 10 (int) value of the "field" field.'

    with pytest.raises(ValueError, match=match(exception_message)):
        instance.other_field = 5

    assert instance.other_field == 30
    assert instance.field == 10


@pytest.mark.parametrize(
    'are_fields_secret',
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    'addictional_arguments',
    [
        {},
        {'doc': 'some doc'},
        {'reverse_conflicts': True},
        {'reverse_conflicts': True, 'doc': 'some doc'},
    ],
)
def test_conflicting_fields_reverse_when_its_on_and_when_set_in_init(addictional_arguments, are_fields_secret):
    doc = addictional_arguments.pop('doc', None)

    class SomeClass(Storage):
        field: int = Field(10, conflicts={'other_field': lambda old, new, other_old, other_new: new > other_new}, doc=doc, secret=are_fields_secret, **addictional_arguments)  # noqa: ARG005
        other_field: int = Field(20, doc=doc, secret=are_fields_secret, **addictional_arguments)

    instance = SomeClass()

    assert instance.field == 10
    assert instance.other_field == 20

    instance = SomeClass(other_field=30)

    assert instance.other_field == 30

    if doc is not None:
        if are_fields_secret:
            exception_message = 'The new *** (int) value of the "other_field" field (some doc) conflicts with the *** (int) value of the "field" field (some doc).'
        else:
            exception_message = 'The new 5 (int) value of the "other_field" field (some doc) conflicts with the 10 (int) value of the "field" field (some doc).'
    elif are_fields_secret:
        exception_message = 'The new *** (int) value of the "other_field" field conflicts with the *** (int) value of the "field" field.'
    else:
        exception_message = 'The new 5 (int) value of the "other_field" field conflicts with the 10 (int) value of the "field" field.'

    with pytest.raises(ValueError, match=match(exception_message)):
        SomeClass(other_field=5)


@pytest.mark.parametrize(
    'reverse_check_parameters',
    [
        {'class': False, 'field': True},
        {'class': True, 'field': False},
        {'class': False, 'field': False},
    ],
)
@pytest.mark.parametrize(
    'are_fields_secret',
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    'addictional_arguments',
    [
        {},
        {'doc': 'some doc'},
    ],
)
def test_basic_conflicting_fields_reverse_when_its_off(addictional_arguments, are_fields_secret, reverse_check_parameters):
    doc = addictional_arguments.pop('doc', None)

    class SomeClass(Storage, reverse_conflicts=reverse_check_parameters['class']):
        field: int = Field(10, conflicts={'other_field': lambda old, new, other_old, other_new: new > other_new}, doc=doc, secret=are_fields_secret, **addictional_arguments, reverse_conflicts=reverse_check_parameters['field'])  # noqa: ARG005
        other_field: int = Field(20, doc=doc, secret=are_fields_secret, **addictional_arguments)

    instance = SomeClass()

    assert instance.field == 10
    assert instance.other_field == 20

    instance.other_field = 30

    assert instance.other_field == 30

    instance.other_field = 5

    assert instance.other_field == 5
    assert instance.field == 10


@pytest.mark.parametrize(
    'reverse_check_parameters',
    [
        {'class': False, 'field': True},
        {'class': True, 'field': False},
        {'class': False, 'field': False},
    ],
)
@pytest.mark.parametrize(
    'are_fields_secret',
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    'addictional_arguments',
    [
        {},
        {'doc': 'some doc'},
    ],
)
def test_conflicting_fields_reverse_when_its_off_and_when_set_in_init(addictional_arguments, are_fields_secret, reverse_check_parameters):
    doc = addictional_arguments.pop('doc', None)

    class SomeClass(Storage, reverse_conflicts=reverse_check_parameters['class']):
        field: int = Field(10, conflicts={'other_field': lambda old, new, other_old, other_new: new > other_new}, doc=doc, secret=are_fields_secret, **addictional_arguments, reverse_conflicts=reverse_check_parameters['field'])  # noqa: ARG005
        other_field: int = Field(20, doc=doc, secret=are_fields_secret, **addictional_arguments)

    instance = SomeClass()

    assert instance.field == 10
    assert instance.other_field == 20

    instance = SomeClass(other_field=30)

    assert instance.other_field == 30

    instance = SomeClass(other_field=5)

    assert instance.other_field == 5
    assert instance.field == 10


@pytest.mark.parametrize(
    'main_field_is_secret',
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    'addictional_arguments',
    [
        {},
        {'doc': 'some doc'},
    ],
)
@pytest.mark.parametrize(
    'reverse_check_parameters',
    [
        {'class': False, 'field': True},
        {'class': True, 'field': False},
        {'class': False, 'field': False},
    ],
)
def test_conflicting_fields_when_reverse_check_off(addictional_arguments, main_field_is_secret, reverse_check_parameters):
    class SomeClass(Storage, reverse_conflicts=reverse_check_parameters['class']):
        field: int = Field(10, conflicts={'other_field': lambda old, new, other_old, other_new: old > other_new}, doc=addictional_arguments.get('doc'), secret=main_field_is_secret, reverse_conflicts=reverse_check_parameters['field'])  # noqa: ARG005
        other_field: int = Field(20, doc=addictional_arguments.get('doc'))

    instance = SomeClass()

    assert instance.field == 10

    instance.other_field = 5

    assert instance.other_field == 5


@pytest.mark.parametrize(
    'main_field_is_secret',
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    'addictional_arguments',
    [
        {},
        {'doc': 'some doc'},
    ],
)
@pytest.mark.parametrize(
    'reverse_check_parameters',
    [
        {'class': False, 'field': True},
        {'class': True, 'field': False},
        {'class': False, 'field': False},
    ],
)
def test_conflicting_fields_in_init_when_reverse_check_off(addictional_arguments, main_field_is_secret, reverse_check_parameters):
    class SomeClass(Storage, reverse_conflicts=reverse_check_parameters['class']):
        field: int = Field(10, conflicts={'other_field': lambda old, new, other_old, other_new: old > other_new}, doc=addictional_arguments.get('doc'), secret=main_field_is_secret, reverse_conflicts=reverse_check_parameters['field'])  # noqa: ARG005
        other_field: int = Field(20, doc=addictional_arguments.get('doc'))

    instance = SomeClass(other_field=5)

    assert instance.field == 10
    assert instance.other_field == 5


@pytest.mark.parametrize(
    'main_field_is_secret',
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    'addictional_arguments',
    [
        {},
        {'doc': 'some doc'},
    ],
)
@pytest.mark.parametrize(
    'reverse_check_parameters',
    [
        {'class': False, 'field': True},
        {'class': True, 'field': False},
        {'class': False, 'field': False},
    ],
)
def test_conflicting_fields_in_defaults_when_reverse_check_off(addictional_arguments, main_field_is_secret, reverse_check_parameters):
    class SomeClass(Storage, reverse_conflicts=reverse_check_parameters['class']):
        field: int = Field(10, conflicts={'other_field': lambda old, new, other_old, other_new: old > other_new}, doc=addictional_arguments.get('doc'), secret=main_field_is_secret, reverse_conflicts=reverse_check_parameters['field'])  # noqa: ARG005
        other_field: int = Field(5, doc=addictional_arguments.get('doc'))

    instance = SomeClass()

    assert instance.field == 10
    assert instance.other_field == 5


def test_variables_order_when_conflicts_checking():
    breadcrumbs = []

    def check_conflicts(old, new, other_old, other_new):
        breadcrumbs.append((old, new, other_old, other_new))
        return old > other_new

    class SomeClass(Storage):
        field: int = Field(5, conflicts={'other_field': check_conflicts})
        other_field: int = Field(10)

    assert len(breadcrumbs) == 1
    assert breadcrumbs[0] == (5, 5, 10, 10)

    instance = SomeClass()

    assert len(breadcrumbs) == 1

    instance.field = 5

    assert len(breadcrumbs) == 2
    assert breadcrumbs[1] == (5, 5, 10, 10)

    instance.field = 6

    assert len(breadcrumbs) == 3
    assert breadcrumbs[2] == (5, 6, 10, 10)

    instance.other_field = 11

    assert len(breadcrumbs) == 4
    assert breadcrumbs[3] == (6, 6, 10, 11)

    instance.other_field = 11

    assert len(breadcrumbs) == 5
    assert breadcrumbs[4] == (6, 6, 11, 11)


def test_there_is_no_dunder_starting_fields_except_user_ones():
    class EmptyClass(Storage):
        ...

    for field_name in dir(EmptyClass()):
        assert field_name.startswith('_')

    class NotEmptyClass(Storage):
        field = Field(5)
        other_field = Field(10)

    for field_name in dir(NotEmptyClass()):
        if field_name not in ('field', 'other_field'):
            assert field_name.startswith('_')


def test_reverse_fields_container_in_basic_case():
    class SomeClass(Storage):
        field: int = Field(5, conflicts={'other_field': lambda old, new, other_old, other_new: old > other_new})  # noqa: ARG005
        other_field: int = Field(10)

    assert SomeClass.__reverse_conflicts__ == {'other_field': ['field']}
    assert SomeClass.__field_names__ == ['field', 'other_field']


def test_reverse_fields_container_in_case_of_inheritance_with_new_field():
    class SomeClass(Storage):
        field: int = Field(5, conflicts={'other_field': lambda old, new, other_old, other_new: old > other_new})  # noqa: ARG005
        other_field: int = Field(10)

    class SomeOtherClass(SomeClass):
        third_field: int = Field(10, conflicts={'other_field': lambda old, new, other_old, other_new: old > 1000})  # noqa: ARG005

    assert SomeClass.__reverse_conflicts__ == {'other_field': ['field']}
    assert SomeOtherClass.__reverse_conflicts__ == {'other_field': ['field', 'third_field']}

    assert SomeClass.__field_names__ == ['field', 'other_field']
    assert SomeOtherClass.__field_names__ == ['field', 'other_field', 'third_field']


def test_reverse_fields_container_in_case_of_inheritance_with_same_field():
    class SomeClass(Storage):
        field: int = Field(5, conflicts={'other_field': lambda old, new, other_old, other_new: old > other_new})  # noqa: ARG005
        other_field: int = Field(10)

    class SomeOtherClass(SomeClass):
        other_field: int = Field(10, conflicts={'field': lambda old, new, other_old, other_new: old > 1000})  # noqa: ARG005

    assert SomeClass.__reverse_conflicts__ == {'other_field': ['field']}
    assert SomeOtherClass.__reverse_conflicts__ == {'other_field': ['field'], 'field': ['other_field']}

    assert SomeClass.__field_names__ == ['field', 'other_field']
    assert SomeOtherClass.__field_names__ == ['field', 'other_field']


@pytest.mark.parametrize(
    'sources',
    [
        [],
        [MemorySource({})],
    ],
)
def test_empty_set_of_sources(sources):
    class SomeClass(Storage, sources=sources):
        field: int = Field(5)
        other_field: int = Field(10)

    instance = SomeClass()

    assert instance.field == 5
    assert instance.other_field == 10


def test_reset_value_using_source():
    class SomeClass(Storage, sources=[MemorySource({'field': 15})]):
        field: int = Field(5)
        other_field: int = Field(10)

    instance = SomeClass()

    assert instance.field == 15
    assert instance.other_field == 10

    instance.field = 7

    assert instance.field == 7


def test_order_of_sources():
    class SomeClass(Storage, sources=[MemorySource({'field': 15}), MemorySource({'field': 23})]):
        field: int = Field(5)
        other_field: int = Field(10)

    instance = SomeClass()

    assert instance.field == 15
    assert instance.other_field == 10

    instance.field = 7

    assert instance.field == 7


@pytest.mark.parametrize(
    'data',
    [
        {'field': 1, 'other_field': 14},
    ],
)
def test_load_from_toml(toml_config_path):
    class SomeClass(Storage, sources=[TOMLSource(toml_config_path)]):
        field: int = Field(5)
        other_field: int = Field(10)

    instance = SomeClass()

    assert instance.field == 1
    assert instance.other_field == 14

    instance.field = 7

    assert instance.field == 7
    assert instance.other_field == 14


@pytest.mark.parametrize(
    'data',
    [
        {'field': 1, 'other_field': 14},
    ],
)
def test_load_from_yaml(yaml_config_path):
    class SomeClass(Storage, sources=[YAMLSource(yaml_config_path)]):
        field: int = Field(5)
        other_field: int = Field(10)

    instance = SomeClass()

    assert instance.field == 1
    assert instance.other_field == 14

    instance.field = 7

    assert instance.field == 7
    assert instance.other_field == 14


@pytest.mark.parametrize(
    'data',
    [
        {'field': 1, 'other_field': 14},
    ],
)
def test_load_from_json(json_config_path):
    class SomeClass(Storage, sources=[JSONSource(json_config_path)]):
        field: int = Field(5)
        other_field: int = Field(10)

    instance = SomeClass()

    assert instance.field == 1
    assert instance.other_field == 14

    instance.field = 7

    assert instance.field == 7
    assert instance.other_field == 14


def test_source_check_is_in_init():
    keys = []

    class PseudoDict:
        def __getitem__(self, key: str) -> Any:
            keys.append(key)
            return 1

    class SomeClass(Storage, sources=[MemorySource(PseudoDict())]):
        field: int = Field(10, read_lock=True)
        other_field: int = Field(20)

    assert not keys

    instance = SomeClass()

    assert keys == ['field', 'other_field']

    assert instance.field == 1
    assert instance.other_field == 1

    assert keys == ['field', 'other_field']


def test_velue_reading_is_under_field_lock_when_its_on():
    locks: List[LockTraceWrapper] = []

    class PseudoDict:
        def get(self, key: str) -> Any:  # noqa: ARG002
            for lock in locks:
                lock.notify('get')
            return 1

    class SomeClass(Storage):
        field: int = Field(10, read_lock=True)
        other_field: int = Field(20)

    instance = SomeClass()

    lock = LockTraceWrapper(instance.__locks__['field'])
    locks.append(lock)
    instance.__locks__['field'] = lock
    instance.__values__ = PseudoDict()

    assert instance.field == 1

    assert lock.was_event_locked('get')


@pytest.mark.parametrize(
    'data',
    [
        {'field': '1'},
    ],
)
def test_read_bad_typed_value_from_toml_source_for_not_deferred_value(toml_config_path):
    class SomeClass(Storage, sources=[TOMLSource(toml_config_path)]):
        field: int = Field(5)

    with pytest.raises(TypeError, match=match('The value of the "field" field did not pass the type check.')):
        SomeClass()


@pytest.mark.parametrize(
    'data',
    [
        {'field': '1'},
    ],
)
def test_read_bad_typed_value_from_yaml_source_for_not_deferred_value(yaml_config_path):
    class SomeClass(Storage, sources=[YAMLSource(yaml_config_path)]):
        field: int = Field(5)

    with pytest.raises(TypeError, match=match('The value of the "field" field did not pass the type check.')):
        SomeClass()


@pytest.mark.parametrize(
    'data',
    [
        {'field': '1'},
    ],
)
def test_read_bad_typed_value_from_json_source_for_not_deferred_value(json_config_path):
    class SomeClass(Storage, sources=[JSONSource(json_config_path)]):
        field: int = Field(5)

    with pytest.raises(TypeError, match=match('The value of the "field" field did not pass the type check.')):
        SomeClass()


@pytest.mark.parametrize(
    'data',
    [
        {'field': [14]},
    ],
)
def test_read_bad_typed_value_from_toml_source_for_deferred_value(toml_config_path):
    class SomeClass(Storage, sources=[TOMLSource(toml_config_path)]):
        field: List[str] = Field(default_factory=list)

    with pytest.raises(TypeError, match=match('The value of the "field" field did not pass the type check.')):
        SomeClass()


@pytest.mark.parametrize(
    'data',
    [
        {'field': [14]},
    ],
)
def test_read_bad_typed_value_from_yaml_source_for_deferred_value(yaml_config_path):
    class SomeClass(Storage, sources=[YAMLSource(yaml_config_path)]):
        field: List[str] = Field(default_factory=list)

    with pytest.raises(TypeError, match=match('The value of the "field" field did not pass the type check.')):
        SomeClass()


@pytest.mark.parametrize(
    'data',
    [
        {'field': [14]},
    ],
)
def test_read_bad_typed_value_from_json_source_for_deferred_value(json_config_path):
    class SomeClass(Storage, sources=[JSONSource(json_config_path)]):
        field: List[str] = Field(default_factory=list)

    with pytest.raises(TypeError, match=match('The value of the "field" field did not pass the type check.')):
        SomeClass()


def test_type_check_with_supertypes():
    class SomeClass(Storage):
        field: NaturalNumber = Field(5)
        other_field: NonNegativeInt = Field(11)

    instance = SomeClass()

    instance.field = 1
    assert instance.field == 1

    instance.field = 1000
    assert instance.field == 1000

    with pytest.raises(TypeError, match=match('The value 0 (int) of the "field" field does not match the type NaturalNumber.')):
        instance.field = 0

    with pytest.raises(TypeError, match=match('The value -1 (int) of the "field" field does not match the type NaturalNumber.')):
        instance.field = -1

    with pytest.raises(TypeError, match=match('The value \'kek\' (str) of the "field" field does not match the type NaturalNumber.')):
        instance.field = 'kek'

    assert instance.field == 1000

    instance.other_field = 1000
    assert instance.other_field == 1000

    instance.other_field = 0
    assert instance.other_field == 0

    with pytest.raises(TypeError, match=match('The value -1 (int) of the "other_field" field does not match the type NonNegativeInt.')):
        instance.other_field = -1

    with pytest.raises(TypeError, match=match('The value \'kek\' (str) of the "other_field" field does not match the type NonNegativeInt.')):
        instance.other_field = 'kek'


def test_wrong_defaults():
    with pytest.raises(ValueError, match=match('You can define a default value or a factory for default values, but not all at the same time.')):
        class SomeClass(Storage):
            field: List[str] = Field([], default_factory=list)


def test_default_value_from_factory():
    class SomeClass(Storage):
        field: List[str] = Field(default_factory=list)

    instance_1 = SomeClass()

    assert instance_1.field == []

    this_field = instance_1.field
    assert instance_1.field is this_field

    instance_2 = SomeClass()

    assert instance_2.field == []

    assert instance_1.field is not instance_2.field

    instance_1.field.append('kek')

    assert instance_1.field == ['kek']
    assert instance_2.field == []

    instance_1.field.append('lol')

    assert instance_1.field == ['kek', 'lol']
    assert instance_2.field == []


def test_type_check_for_default_factory():
    class SomeClass(Storage):
        field: int = Field(default_factory=lambda: 'kek')

    with pytest.raises(TypeError, match=match('The value \'kek\' (str) of the "field" field does not match the type int.')):
        SomeClass()


@pytest.mark.parametrize(
    'addictional_parameters',
    [
        {},
        {'validate_default': True},
    ],
)
def test_validate_default_factory_value_fith_function_when_its_on_and_validation_not_passed(addictional_parameters):
    class SomeClass(Storage):
        field: str = Field(default_factory=lambda: 'kek', validation=lambda x: x != 'kek', **addictional_parameters)

    with pytest.raises(ValueError, match=match('The value \'kek\' (str) of the "field" field does not match the validation.')):
        SomeClass()


@pytest.mark.parametrize(
    'addictional_parameters',
    [
        {},
        {'validate_default': True},
    ],
)
def test_validate_default_factory_value_fith_function_when_its_on_and_validation_passed(addictional_parameters):
    class SomeClass(Storage):
        field: str = Field(default_factory=lambda: 'kek', validation=lambda x: x == 'kek', **addictional_parameters)

    instance = SomeClass()

    assert instance.field == 'kek'


@pytest.mark.parametrize(
    'addictional_parameters',
    [
        {},
        {'validate_default': True},
    ],
)
def test_validate_default_factory_value_fith_dict_when_its_on_and_validation_not_passed(addictional_parameters):
    class SomeClass(Storage):
        field: str = Field(default_factory=lambda: 'kek', validation={'some message': lambda x: x != 'kek'}, **addictional_parameters)

    with pytest.raises(ValueError, match=match('some message')):
        SomeClass()


@pytest.mark.parametrize(
    'addictional_parameters',
    [
        {},
        {'validate_default': True},
    ],
)
def test_validate_default_factory_value_fith_dict_when_its_on_and_validation_passed(addictional_parameters):
    class SomeClass(Storage):
        field: str = Field(default_factory=lambda: 'kek', validation={'some message': lambda x: x == 'kek'}, **addictional_parameters)

    instance = SomeClass()

    assert instance.field == 'kek'


def test_validate_default_factory_value_fith_function_when_its_off_and_validation_not_passed():
    class SomeClass(Storage):
        field: str = Field(default_factory=lambda: 'kek', validation=lambda x: x != 'kek', validate_default=False)

    instance = SomeClass()

    assert instance.field == 'kek'


def test_validate_default_factory_value_fith_function_when_its_off_and_validation_passed():
    class SomeClass(Storage):
        field: str = Field(default_factory=lambda: 'kek', validation=lambda x: x == 'kek', validate_default=False)

    instance = SomeClass()

    assert instance.field == 'kek'


def test_validate_default_factory_value_fith_dict_when_its_off_and_validation_not_passed():
    class SomeClass(Storage):
        field: str = Field(default_factory=lambda: 'kek', validation={'some message': lambda x: x != 'kek'}, validate_default=False)

    instance = SomeClass()

    assert instance.field == 'kek'


def test_validate_default_factory_value_fith_dict_when_its_off_and_validation_passed():
    class SomeClass(Storage):
        field: str = Field(default_factory=lambda: 'kek', validation={'some message': lambda x: x == 'kek'}, validate_default=False)

    instance = SomeClass()

    assert instance.field == 'kek'


def test_conflicts_for_default_factory():
    field_value = 10
    other_lazy_field_value = 5
    class SomeClass(Storage):
        field: int = Field(default_factory=lambda: field_value, conflicts={'other_field': lambda old, new, other_old, other_new: new > other_old, 'other_lazy_field': lambda old, new, other_old, other_new: new > other_old})  # noqa: ARG005
        other_field: int = Field(20)
        other_lazy_field: int = Field(default_factory=lambda: other_lazy_field_value)

    with pytest.raises(ValueError, match=match('The 10 (int) deferred default value of the "field" field conflicts with the 5 (int) value of the "other_lazy_field" field.')):
        SomeClass()

    field_value = 25

    with pytest.raises(ValueError, match=match('The 25 (int) deferred default value of the "field" field conflicts with the 20 (int) value of the "other_field" field.')):
        SomeClass()

    field_value = 5
    other_lazy_field_value = 30

    instance = SomeClass()

    assert instance.field == 5
    assert instance.other_field == 20
    assert instance.other_lazy_field == 30


def test_reverse_conflicts_for_default_factory():
    other_lazy_field_value = 5

    class SomeClass(Storage):
        field: int = Field(10, conflicts={'other_lazy_field': lambda old, new, other_old, other_new: new > other_old})  # noqa: ARG005
        other_lazy_field: int = Field(default_factory=lambda: other_lazy_field_value)

    with pytest.raises(ValueError, match=match('The 10 (int) deferred default value of the "field" field conflicts with the 5 (int) value of the "other_lazy_field" field.')):
        SomeClass()

    other_lazy_field_value = 15

    instance = SomeClass()

    assert instance.field == 10
    assert instance.other_lazy_field == 15


@pytest.mark.parametrize(
    ('class_flag', 'field_flag'),
    [
        (True, False),
        (False, True),
        (False, False),
    ],
)
def test_reverse_conflicts_off_for_default_factory(class_flag, field_flag):
    other_lazy_field_value = 5

    class SomeClass(Storage, reverse_conflicts=class_flag):
        field: int = Field(10, conflicts={'other_lazy_field': lambda old, new, other_old, other_new: new > other_old}, reverse_conflicts=field_flag)  # noqa: ARG005
        other_lazy_field: int = Field(default_factory=lambda: other_lazy_field_value)

    instance = SomeClass()

    assert instance.field == 10
    assert instance.other_lazy_field == 5

    other_lazy_field_value = 15

    instance = SomeClass()

    assert instance.field == 10
    assert instance.other_lazy_field == 15


def test_conversion_is_not_under_field_lock():
    locks = []

    def conversion(value: int) -> int:
        for lock in locks:
            lock.notify('conversion')
        return value * 2

    class SomeClass(Storage):
        field = Field(42, conversion=conversion)

    storage = SomeClass()

    lock = LockTraceWrapper(storage.__locks__['field'])
    storage.__locks__['field'] = lock
    locks.append(lock)

    storage.field = 5

    assert storage.field == 10

    assert not lock.was_event_locked('conversion')
    assert lock.trace


def test_conflicts_check_on_set_is_after_conversion():
    class SomeClass(Storage):
        field: int = Field(5, conversion=lambda x: x * 2, conflicts={'other_field': lambda old, new, other_old, other_new: new > other_new})  # noqa: ARG005
        other_field: int = Field(10)

    instance = SomeClass()

    with pytest.raises(ValueError, match=match('The new 20 (int) value of the "field" field conflicts with the 10 (int) value of the "other_field" field.')):
        instance.field = 10


def test_conflicts_check_on_defaults_is_after_conversion():
    with pytest.raises(ValueError, match=match('The 20 (int) default value of the "field" field conflicts with the 10 (int) value of the "other_field" field.')):
        class SomeClass(Storage):
            field: int = Field(10, conversion=lambda x: x * 2, conflicts={'other_field': lambda old, new, other_old, other_new: new > other_new})  # noqa: ARG005
            other_field: int = Field(10)


def test_value_check_for_defaults_is_after_conversion():
    with pytest.raises(ValueError, match=match('The value 20 (int) of the "field" field does not match the validation.')):
        class SomeClass(Storage):
            field: int = Field(10, conversion=lambda x: x * 2, validation=lambda x: x == 10)
            other_field: int = Field(10)


def test_value_check_for_set_is_after_conversion():
    class SomeClass(Storage):
        field: int = Field(5, conversion=lambda x: x * 2, validation=lambda x: x == 10)
        other_field: int = Field(10)

    instance = SomeClass()

    with pytest.raises(ValueError, match=match('The value 20 (int) of the "field" field does not match the validation.')):
        instance.field = 10


def test_type_check_for_defaults_is_before_conversion():
    with pytest.raises(TypeError, match=match('The value 5 (int) of the "field" field does not match the type str.')):
        class SomeClass(Storage):
            field: str = Field(5, conversion=lambda x: str(x))


def test_type_check_for_defaults_is_after_conversion():
    with pytest.raises(TypeError, match=match('The value \'5\' (str) of the "field" field does not match the type int.')):
        class SomeClass(Storage):
            field: int = Field(5, conversion=lambda x: str(x))


def test_type_check_for_set_is_before_conversion():
    class SomeClass(Storage):
        field: Union[int, str] = Field(5, conversion=lambda x: str(x))

    instance = SomeClass()

    assert instance.field == '5'

    if sys.version_info < (3, 10):
        type_representation = 'typing.Union'
    else:
        type_representation = 'Union'

    with pytest.raises(TypeError, match=match(f'The value 5.5 (float) of the "field" field does not match the type {type_representation}.')):
        instance.field = 5.5


def test_type_check_for_set_is_after_conversion():
    class SomeClass(Storage):
        field: int = Field(5, conversion=lambda x: x if x == 5 else str(x))

    instance = SomeClass()

    assert instance.field == 5

    with pytest.raises(TypeError, match=match('The value \'6\' (str) of the "field" field does not match the type int.')):
        instance.field = 6


def test_basic_conversion_when_set_and_init_with_passed_type_check_for_new_and_old_results():
    class SomeClass(Storage):
        field: int = Field(10, conversion=lambda x: x * 2)

    instance = SomeClass()

    assert instance.field == 20

    instance.field = 3

    assert instance.field == 6


@pytest.mark.parametrize(
    'data',
    [
        {'field': 15},
    ],
)
def test_conversion_for_source(toml_config_path, json_config_path, yaml_config_path):
    class SomeClass(Storage, sources=[TOMLSource(toml_config_path)]):
        field: int = Field(10, conversion=lambda x: x * 2)

    assert SomeClass().field == 30

    class SecondClass(Storage, sources=[JSONSource(json_config_path)]):
        field: int = Field(10, conversion=lambda x: x * 2)

    assert SecondClass().field == 30

    class SecondClass(Storage, sources=[YAMLSource(yaml_config_path)]):
        field: int = Field(10, conversion=lambda x: x * 2)

    assert SecondClass().field == 30


@pytest.mark.parametrize(
    'data',
    [
        {'field': 15},
    ],
)
def test_type_check_before_conversion_for_toml_source(toml_config_path):
    class SomeClass(Storage, sources=[TOMLSource(toml_config_path)]):
        field: str = Field('kek', conversion=lambda x: str(x))

    with pytest.raises(TypeError, match=match('The value of the "field" field did not pass the type check.')):
        SomeClass()


@pytest.mark.parametrize(
    'data',
    [
        {'field': 15},
    ],
)
def test_type_check_before_conversion_for_yaml_source(yaml_config_path):
    class SomeClass(Storage, sources=[YAMLSource(yaml_config_path)]):
        field: str = Field('kek', conversion=lambda x: str(x))

    with pytest.raises(TypeError, match=match('The value of the "field" field did not pass the type check.')):
        SomeClass()


@pytest.mark.parametrize(
    'data',
    [
        {'field': 15},
    ],
)
def test_type_check_before_conversion_for_json_source(json_config_path):
    class SomeClass(Storage, sources=[JSONSource(json_config_path)]):
        field: str = Field('kek', conversion=lambda x: str(x))

    with pytest.raises(TypeError, match=match('The value of the "field" field did not pass the type check.')):
        SomeClass()


@pytest.mark.parametrize(
    'data',
    [
        {'field': 15},
    ],
)
def test_type_check_after_conversion_for_source(toml_config_path, json_config_path, yaml_config_path):
    with pytest.raises(TypeError, match=match('The value \'10\' (str) of the "field" field does not match the type int.')):
        class SomeClass(Storage, sources=[TOMLSource(toml_config_path)]):
            field: int = Field(10, conversion=lambda x: str(x))

    with pytest.raises(TypeError, match=match('The value \'10\' (str) of the "field" field does not match the type int.')):
        class SomeClass(Storage, sources=[JSONSource(json_config_path)]):
            field: int = Field(10, conversion=lambda x: str(x))

    with pytest.raises(TypeError, match=match('The value \'10\' (str) of the "field" field does not match the type int.')):
        class SomeClass(Storage, sources=[YAMLSource(yaml_config_path)]):
            field: int = Field(10, conversion=lambda x: str(x))


def test_conversion_for_default_factory():
    class SomeClass(Storage):
        field: int = Field(default_factory=lambda: 10, conversion=lambda x: x * 2)

    assert SomeClass().field == 20


def test_type_check_is_before_conversion_for_default_factory():
    class SomeClass(Storage):
        field: str = Field(default_factory=lambda: 10, conversion=lambda x: str(x))

    with pytest.raises(TypeError, match=match('The value 10 (int) of the "field" field does not match the type str.')):
        SomeClass()


def test_type_check_is_after_conversion_for_default_factory():
    class SomeClass(Storage):
        field: int = Field(default_factory=lambda: 10, conversion=lambda x: str(x))

    with pytest.raises(TypeError, match=match('The value \'10\' (str) of the "field" field does not match the type int.')):
        SomeClass()


def test_validation_is_after_conversion_for_default_factory():
    class SomeClass(Storage):
        field: int = Field(default_factory=lambda: 5, conversion=lambda x: 10, validation=lambda x: x != 10)  # noqa: ARG005

    with pytest.raises(ValueError, match=match('The value 10 (int) of the "field" field does not match the validation.')):
        SomeClass()


def test_validation_is_after_conversion_for_default_factory_when_its_off():
    class SomeClass(Storage):
        field: int = Field(default_factory=lambda: 5, conversion=lambda x: 10, validation=lambda x: x != 10, validate_default=False)  # noqa: ARG005

    instance = SomeClass()

    assert instance.field == 10

    with pytest.raises(ValueError, match=match('The value 10 (int) of the "field" field does not match the validation.')):
        instance.field = 5


def test_share_locks():
    class SomeClass(Storage):
        first_field: int = Field(1, share_mutex_with=['second_field'])
        second_field: int = Field(2)
        third_field: int = Field(3)
        forth_field: int = Field(4, conflicts={'fifth_field': lambda x, y, z, m: False})  # noqa: ARG005
        fifth_field: int = Field(5)

    instance = SomeClass()

    assert instance.__locks__['first_field'] is instance.__locks__['second_field']

    assert instance.__locks__['first_field'] is not instance.__locks__['third_field']
    assert instance.__locks__['first_field'] is not instance.__locks__['forth_field']
    assert instance.__locks__['first_field'] is not instance.__locks__['fifth_field']

    assert instance.__locks__['second_field'] is not instance.__locks__['third_field']
    assert instance.__locks__['second_field'] is not instance.__locks__['forth_field']
    assert instance.__locks__['second_field'] is not instance.__locks__['fifth_field']

    assert instance.__locks__['third_field'] is not instance.__locks__['forth_field']
    assert instance.__locks__['third_field'] is not instance.__locks__['fifth_field']

    assert instance.__locks__['forth_field'] is instance.__locks__['fifth_field']


def test_non_existing_field_to_share_mutex():
    with pytest.raises(NameError, match=match('You indicated that you need to share the mutex of "first_field" field with field "sacond_field", but field "sacond_field" does not exist.')):
        class SomeClass(Storage):
            first_field: int = Field(1, share_mutex_with=['sacond_field'])
            second_field: int = Field(2)


def test_share_mutex_with_twice():
    class SomeClass(Storage):
        first_field: int = Field(1, share_mutex_with=['second_field'])
        second_field: int = Field(2, share_mutex_with=['third_field'])
        third_field: int = Field(3)

    instance = SomeClass()

    assert instance.__locks__['first_field'] is instance.__locks__['second_field']
    assert instance.__locks__['third_field'] is instance.__locks__['third_field']


def test_share_mutex_with_conflicting_field():
    class SomeClass(Storage):
        first_field: int = Field(1, share_mutex_with=['second_field'], conflicts={'third_field': lambda x, y, z, m: False})  # noqa: ARG005
        second_field: int = Field(2)
        third_field: int = Field(3)

    instance = SomeClass()

    assert instance.__locks__['first_field'] is instance.__locks__['second_field']
    assert instance.__locks__['third_field'] is instance.__locks__['third_field']


def test_get_something_from_env(monkeypatch):
    monkeypatch.setenv("SKELET_FIELD", "1")
    monkeypatch.setenv("SKELET_ANOTHER_FIELD", "kek")

    class SomeClass(Storage, sources=EnvSource.for_library('skelet')):
        field: int = Field(10)
        another_field: str = Field('lol')
        third_field: List[int] = Field(default_factory=lambda: [1, 2, 3])

    instance = SomeClass()

    assert instance.field == 1
    assert instance.another_field == 'kek'
    assert instance.third_field == [1, 2, 3]


def test_get_value_from_sources_by_aliases():
    class SomeClass(Storage, sources=[MemorySource({'field': 1, 'a-b-c': 2})]):
        first_field: int = Field(123, alias='field')
        second_field: int = Field(456, alias='a-b-c')

    instance = SomeClass()

    assert instance.first_field == 1
    assert instance.second_field == 2


def test_get_value_from_sources_by_aliases_when_there_are_original_field_names_available():
    class SomeClass(Storage, sources=[MemorySource({'field': 1, 'a-b-c': 2, 'first_field': 3, 'second_field': 4})]):
        first_field: int = Field(123, alias='field')
        second_field: int = Field(456, alias='a-b-c')

    instance = SomeClass()

    assert instance.first_field == 1
    assert instance.second_field == 2


def test_per_field_sources():
    class SomeClass(Storage):
        first_field: int = Field(123, sources=[MemorySource({'first_field': 1, 'second_field': 2})])
        second_field: int = Field(456, sources=[MemorySource({'first_field': 1, 'second_field': 2})])

    instance = SomeClass()

    assert instance.first_field == 1
    assert instance.second_field == 2


def test_per_field_sources_in_conflict_with_class_source():
    class SomeClass(Storage, sources=[MemorySource({'first_field': 4, 'second_field': 5})]):
        first_field: int = Field(123, sources=[MemorySource({'first_field': 1, 'second_field': 2})])
        second_field: int = Field(456, sources=[MemorySource({'first_field': 1})])

    instance = SomeClass()

    assert instance.first_field == 1
    assert instance.second_field == 456


def test_per_field_sources_with_ellipsis_in_conflict_with_class_source():
    class SomeClass(Storage, sources=[MemorySource({'first_field': 4, 'second_field': 5})]):
        first_field: int = Field(123, sources=[MemorySource({'first_field': 1, 'second_field': 2}), ...])
        second_field: int = Field(456, sources=[MemorySource({'first_field': 1}), ...])

    instance = SomeClass()

    assert instance.first_field == 1
    assert instance.second_field == 5


def test_default_value_is_not_set():
    class SomeClass(Storage):
        first_field: int = Field()
        second_field: int = Field()

    with pytest.raises(ValueError, match=match('The value for the "first_field" field is undefined. Set the default value, or specify the value when creating the instance.')):
        SomeClass()

    with pytest.raises(ValueError, match=match('The value for the "second_field" field is undefined. Set the default value, or specify the value when creating the instance.')):
        SomeClass(first_field=5)

    instance = SomeClass(first_field=5, second_field=10)

    assert instance.first_field == 5
    assert instance.second_field == 10



def test_default_value_is_not_set_but_there_is_source():
    class SomeClass(Storage, sources=[MemorySource({'first_field': 4, 'second_field': 5})]):
        first_field: int = Field()
        second_field: int = Field()

    instance = SomeClass()

    assert instance.first_field == 4
    assert instance.second_field == 5


def test_default_value_is_not_set_but_there_are_per_field_sources():
    class SomeClass(Storage):
        first_field: int = Field(sources=[MemorySource({'first_field': 4, 'second_field': 5})])
        second_field: int = Field(sources=[MemorySource({'first_field': 4, 'second_field': 5})])

    instance = SomeClass()

    assert instance.first_field == 4
    assert instance.second_field == 5


def test_wrong_default_factories():
    with pytest.raises(SignatureMismatchError, match=match('The default value factory should not expect any arguments.')):
        class SomeClass(Storage):
            field: int = Field(default_factory=lambda x: x)

    with pytest.raises(SignatureMismatchError, match=match('The default value factory should not expect any arguments.')):
        class SomeClass(Storage):
            field: int = Field(default_factory=lambda x, y: x + y)


def test_wrong_validation_function():
    with pytest.raises(SignatureMismatchError, match=match('A function that accepts only one positional argument is expected as a field validator.')):
        class SomeClass(Storage):
            field: int = Field(123, validation=lambda: False)

    with pytest.raises(SignatureMismatchError, match=match('A function that accepts only one positional argument is expected as a field validator.')):
        class SomeClass(Storage):
            field: int = Field(123, validation=lambda x, y: x + y)

    with pytest.raises(SignatureMismatchError, match=match("Field validator with message 'some message' is incorrect: a function that accepts only one positional argument is expected.")):
        class SomeClass(Storage):
            field: int = Field(123, validation={'some message': lambda: False})

    with pytest.raises(SignatureMismatchError, match=match("Field validator with message 'some message' is incorrect: a function that accepts only one positional argument is expected.")):
        class SomeClass(Storage):
            field: int = Field(123, validation={'some message': lambda x, y: x + y})

    with pytest.raises(SignatureMismatchError, match=match("Field validator with message 'some another message' is incorrect: a function that accepts only one positional argument is expected.")):
        class SomeClass(Storage):
            field: int = Field(123, validation={'some message': lambda x: False, 'some another message': lambda: False})  # noqa: ARG005

    with pytest.raises(SignatureMismatchError, match=match("Field validator with message 'some another message' is incorrect: a function that accepts only one positional argument is expected.")):
        class SomeClass(Storage):
            field: int = Field(123, validation={'some message': lambda x: False, 'some another message': lambda x, y: x + y})  # noqa: ARG005


def test_wrong_conflict_checker():
    with pytest.raises(SignatureMismatchError, match=match("The function for checking conflicts with field 'another_field' is bad; it should take four positional arguments: the old value of this field, the new value of this field, the old value of the conflicting field, and the new value of the conflicting field (for reverse checks).")):
        class SomeClass(Storage):
            field: int = Field(123, conflicts={'another_field': lambda: False})
            another_field: int = Field(123)

    with pytest.raises(SignatureMismatchError, match=match("The function for checking conflicts with field 'another_field' is bad; it should take four positional arguments: the old value of this field, the new value of this field, the old value of the conflicting field, and the new value of the conflicting field (for reverse checks).")):
        class SomeClass(Storage):
            field: int = Field(123, conflicts={'another_field': lambda x: False})  # noqa: ARG005
            another_field: int = Field(123)

    with pytest.raises(SignatureMismatchError, match=match("The function for checking conflicts with field 'another_field' is bad; it should take four positional arguments: the old value of this field, the new value of this field, the old value of the conflicting field, and the new value of the conflicting field (for reverse checks).")):
        class SomeClass(Storage):
            field: int = Field(123, conflicts={'another_field': lambda x, y: False})  # noqa: ARG005
            another_field: int = Field(123)

    with pytest.raises(SignatureMismatchError, match=match("The function for checking conflicts with field 'another_field' is bad; it should take four positional arguments: the old value of this field, the new value of this field, the old value of the conflicting field, and the new value of the conflicting field (for reverse checks).")):
        class SomeClass(Storage):
            field: int = Field(123, conflicts={'another_field': lambda x, y, z: False})  # noqa: ARG005
            another_field: int = Field(123)

    with pytest.raises(SignatureMismatchError, match=match("The function for checking conflicts with field 'another_field' is bad; it should take four positional arguments: the old value of this field, the new value of this field, the old value of the conflicting field, and the new value of the conflicting field (for reverse checks).")):
        class SomeClass(Storage):
            field: int = Field(123, conflicts={'another_field': lambda x, y, z, a, b: False})  # noqa: ARG005
            another_field: int = Field(123)


def test_wrong_converter_function():
    with pytest.raises(SignatureMismatchError, match=match('The value converter must accept only one argument: the value before conversion.')):
        class SomeClass(Storage):
            field: int = Field(123, conversion=lambda: 456)

    with pytest.raises(SignatureMismatchError, match=match('The value converter must accept only one argument: the value before conversion.')):
        class SomeClass(Storage):
            field: int = Field(123, conversion=lambda x, y: 456)  # noqa: ARG005

    with pytest.raises(SignatureMismatchError, match=match('The value converter must accept only one argument: the value before conversion.')):
        class SomeClass(Storage):
            field: int = Field(123, conversion=lambda x, y, z: 456)  # noqa: ARG005


def test_wrong_change_action():
    with pytest.raises(SignatureMismatchError, match=match('The callback for each field change must take 3 arguments: the old field value, the new value, and the storage object itself.')):
        class SomeClass(Storage):
            field: int = Field(123, action=lambda: None)

    with pytest.raises(SignatureMismatchError, match=match('The callback for each field change must take 3 arguments: the old field value, the new value, and the storage object itself.')):
        class SomeClass(Storage):
            field: int = Field(123, action=lambda x: None)  # noqa: ARG005

    with pytest.raises(SignatureMismatchError, match=match('The callback for each field change must take 3 arguments: the old field value, the new value, and the storage object itself.')):
        class SomeClass(Storage):
            field: int = Field(123, action=lambda x, y: None)  # noqa: ARG005

    with pytest.raises(SignatureMismatchError, match=match('The callback for each field change must take 3 arguments: the old field value, the new value, and the storage object itself.')):
        class SomeClass(Storage):
            field: int = Field(123, action=lambda x, y, z, another: None)  # noqa: ARG005



from typing import Any, List, Optional, Type, cast

from denial import InnerNoneType
from printo import repred

from skelet.sources.abstract import AbstractSource, ExpectedType

sentinel = InnerNoneType()

@repred(prefer_positional=True)  # type: ignore[call-overload]
class SourcesCollection(AbstractSource[ExpectedType]):
    def __init__(self, sources: List[AbstractSource[ExpectedType]]) -> None:
        self.sources = sources

    def __getitem__(self, key: str) -> Any:
        for source in self.sources:
            try:
                return source[key]
            except KeyError:
                pass

        raise KeyError(key)

    def get(self, key: str, default: Any = None) -> Any:
        try:
            return self[key]
        except KeyError:
            return default

    def type_awared_get(self, key: str, hint: Type[ExpectedType], default: ExpectedType = cast(ExpectedType, sentinel)) -> Optional[ExpectedType]:  # noqa: B008
        for source in self.sources:
            maybe_result = source.type_awared_get(key, hint, default=default)
            if maybe_result is not default:
                return maybe_result

        if default is not sentinel:
            return default

        return None



from functools import cached_property
from pathlib import Path
from typing import Dict, List, Optional, Union

try:
    # TODO: This appeared in Python 3.11, this condition needs to be removed when there are no earlier versions in CI.
    from tomllib import load  # type: ignore[import-not-found, unused-ignore]
except ImportError:  # pragma: no cover
    from tomli import (  # type: ignore[import-not-found, no-redef, assignment, unused-ignore]
        load,  # type: ignore[assignment, import-not-found, no-redef, unused-ignore]
    )

from printo import repred

from skelet.sources.abstract import AbstractSource, ExpectedType


@repred(positionals=['path'], filters={'allow_non_existent_files': lambda x: x != True, 'table': lambda x: bool(x)}, getters={'table': lambda x: x.table})
class TOMLSource(AbstractSource[ExpectedType]):
    def __init__(self, path: Union[str, Path], table: Optional[Union[str, List[str]]] = None, allow_non_existent_files: bool = True) -> None:
        self.path = path
        self.allow_non_existent_files = allow_non_existent_files

        if isinstance(table, str):
            self.table = table.split('.')
        elif isinstance(table, list):
            self.table = []
            for subtable in table:
                self.table.extend(subtable.split('.'))
        else:
            self.table = []

        for subtable in self.table:
            if not subtable.isidentifier():
                raise ValueError(f'You can only use a subset of all valid TOML format identifiers that can be used as a Python identifier. You used "{subtable}".')

    def __getitem__(self, key: str) -> ExpectedType:
        return self.data[key]

    @cached_property
    def data(self) -> Dict[str, ExpectedType]:
        try:
            with open(self.path, 'rb') as file:
                table = load(file)

            for subtable_name in self.table:
                table = table[subtable_name]

            return table  # type: ignore[no-any-return, unused-ignore]

        except FileNotFoundError:
            if self.allow_non_existent_files:
                return {}
            raise

        except KeyError:
            return {}

    @classmethod
    def for_library(cls, library_name: str) -> List['TOMLSource[ExpectedType]']:
        if not library_name.isidentifier():
            raise ValueError('The library name can only be a valid Python identifier.')

        return [cls(f'{library_name}.toml'), cls(f'.{library_name}.toml'), cls('pyproject.toml', table=f'tool.{library_name}')]



from typing import Dict, List

from printo import repred

from skelet.sources.abstract import AbstractSource, ExpectedType


@repred(prefer_positional=True)  # type: ignore[call-overload]
class MemorySource(AbstractSource[ExpectedType]):
    def __init__(self, data: Dict[str, ExpectedType]) -> None:
        self.data = data

    def __getitem__(self, key: str) -> ExpectedType:
        return self.data[key]

    @classmethod
    def for_library(cls, library_name: str) -> List['MemorySource[ExpectedType]']:
        if not library_name.isidentifier():
            raise ValueError('The library name can only be a valid Python identifier.')
        return []



from collections import defaultdict
from threading import Lock
from typing import Any, Dict, List, Optional, Tuple, Union

from denial import InnerNoneType
from locklib import ContextLockProtocol
from printo import descript_data_object

from skelet.sources.abstract import AbstractSource, ExpectedType
from skelet.sources.collection import SourcesCollection

sentinel = InnerNoneType()

class Storage:
    __values__: Dict[str, Any]
    __locks__: Dict[str, ContextLockProtocol]
    __field_names__: Union[List[str], Tuple[str, ...]] = ()
    __reverse_conflicts__: Dict[str, List[str]]
    __sources__: SourcesCollection  # type: ignore[type-arg]

    def __init__(self, **kwargs: Any) -> None:
        self.__values__: Dict[str, Any] = {}
        self.__locks__ = {field_name: Lock() for field_name in self.__field_names__}
        deduplicated_fields = set(self.__field_names__)

        for field_name in self.__field_names__:
            field = getattr(type(self), field_name)
            lock = self.__locks__[field_name]
            if field.conflicts is not None:
                for another_field_name in field.conflicts:
                    self.__locks__[another_field_name] = lock
            if field.share_mutex_with is not None:
                for another_field_name in field.share_mutex_with:
                    self.__locks__[another_field_name] = lock

        for field_name in self.__field_names__:
            field = getattr(type(self), field_name)
            content = field.get_sources(self).type_awared_get(field.alias, field.type_hint, sentinel)
            it_is_not_default = True
            if content is not sentinel:
                field.check_type_hints(content, strict=True, raise_all=True)
                field.check_value(content, raise_all=True)
            elif field._default_factory is not None:
                content = field._default_factory()
                field.check_type_hints(content, strict=True, raise_all=True)
                if field.validate_default:
                    field.check_value(content, raise_all=True)
            else:
                it_is_not_default = False
                content = field._default

            if field.conversion is not None and it_is_not_default:
                content = field.conversion(content)
                field.check_type_hints(content, strict=True, raise_all=True)
                if field.validate_default:
                    field.check_value(content, raise_all=True)

            self.__values__[field_name] = content

        for field_name in self.__field_names__:
            field = getattr(type(self), field_name)

            if field._default_factory is not None:
                if field.conflicts is not None:
                    for conflicting_field_name, checker in field.conflicts.items():
                        if checker(self.__values__[field_name], self.__values__[field_name], self.__values__[conflicting_field_name], self.__values__[conflicting_field_name]):
                            conflicting_field = getattr(type(self), conflicting_field_name)
                            raise ValueError(f'The {field.get_value_representation(self.__values__[field_name])} deferred default value of the {field.get_field_name_representation()} conflicts with the {conflicting_field.get_value_representation(self.__values__[conflicting_field_name])} value of the {conflicting_field.get_field_name_representation()}.')

                if field_name in self.__reverse_conflicts__:
                    conflicting_field_names = self.__reverse_conflicts__[field_name]
                    for conflicting_field_name in conflicting_field_names:
                        conflicting_field = getattr(type(self), conflicting_field_name)
                        checker = conflicting_field.conflicts[field_name]
                        if checker(self.__values__[conflicting_field_name], self.__values__[conflicting_field_name], self.__values__[field_name], self.__values__[field_name]):
                            raise ValueError(f'The {conflicting_field.get_value_representation(self.__values__[conflicting_field_name])} deferred default value of the {conflicting_field.get_field_name_representation()} conflicts with the {field.get_value_representation(self.__values__[field_name])} value of the {field.get_field_name_representation()}.')

        for key, value in kwargs.items():
            if key not in deduplicated_fields:
                raise KeyError(f'The "{key}" field is not defined.')
            setattr(self, key, value)

        for field_name in self.__field_names__:
            field_content = getattr(self, field_name)
            if isinstance(field_content, InnerNoneType):
                raise ValueError(f'The value for the "{field_name}" field is undefined. Set the default value, or specify the value when creating the instance.')


    def __init_subclass__(cls, reverse_conflicts: bool = True, sources: Optional[List[AbstractSource[ExpectedType]]] = None, **kwargs: Any):
            super().__init_subclass__(**kwargs)

            for field_name in cls.__field_names__:
                field = getattr(cls, field_name)
                if field.exception is not None:
                    raise field.exception

            cls.__sources__ = SourcesCollection(sources) if sources is not None else SourcesCollection([])

            deduplicated_field_names = set(cls.__field_names__)

            cls.__reverse_conflicts__ = defaultdict(list)
            for field_name in cls.__field_names__:
                field = getattr(cls, field_name)

                if field.conflicts is not None:
                    for other_field_name in field.conflicts:
                        if field.reverse_conflicts_on and reverse_conflicts:
                            cls.__reverse_conflicts__[other_field_name].append(field_name)

            for field_name in cls.__field_names__:
                field = getattr(cls, field_name)

                if field.share_mutex_with is not None:
                    for another_field_name in field.share_mutex_with:
                        if another_field_name not in deduplicated_field_names:
                            raise NameError(f'You indicated that you need to share the mutex of {field.get_field_name_representation()} with field "{another_field_name}", but field "{another_field_name}" does not exist.')

                if field.conflicts is not None:
                    for conficting_field_name, checker in field.conflicts.items():
                        if conficting_field_name not in deduplicated_field_names:
                            raise NameError(f'You have set a conflict condition for {field.get_field_name_representation()} with field "{conficting_field_name}", but the field "{conficting_field_name}" does not exist in the class {cls.__name__}.')
                        if not isinstance(field._default, InnerNoneType) and not isinstance(getattr(cls, conficting_field_name)._default, InnerNoneType) and reverse_conflicts and field.reverse_conflicts_on and checker(field._default, field._default, getattr(cls, conficting_field_name)._default, getattr(cls, conficting_field_name)._default):
                            other_field = getattr(cls, conficting_field_name)
                            raise ValueError(f'The {field.get_value_representation(field._default)} default value of the {field.get_field_name_representation()} conflicts with the {other_field.get_value_representation(other_field._default)} value of the {other_field.get_field_name_representation()}.')

    def __repr__(self) -> str:
        fields_content = {}
        secrets = {}

        for field_name in self.__field_names__:
            fields_content[field_name] = getattr(self, field_name)
            if getattr(type(self), field_name).secret:
                secrets[field_name] = '***'

        return descript_data_object(type(self).__name__, (), fields_content, placeholders=secrets)  # type: ignore[arg-type]



from simtypes import NaturalNumber as NaturalNumber  # noqa: I001
from simtypes import NonNegativeInt as NonNegativeInt

from skelet.fields.base import Field as Field
from skelet.functions.asdict import asdict as asdict
from skelet.sources.cli import FixedCLISource as FixedCLISource
from skelet.sources.env import EnvSource as EnvSource
from skelet.sources.json import JSONSource as JSONSource
from skelet.sources.memory import MemorySource as MemorySource
from skelet.sources.toml import TOMLSource as TOMLSource
from skelet.sources.yaml import YAMLSource as YAMLSource
from skelet.storage import Storage as Storage
from skelet.sources.getter_for_libraries import for_tool as for_tool



import os
from argparse import ArgumentParser
from contextlib import redirect_stderr
from typing import List, Optional, Type, Union, cast

from denial import InnerNoneType, SentinelType
from printo import repred
from simtypes import from_string

from skelet.errors import CLIFormatError
from skelet.sources.abstract import AbstractSource, ExpectedType

sentinel = InnerNoneType()

@repred(
    getters={
        'positional_arguments': lambda x: x.position_arguments,
        'named_arguments': lambda x: x.named_arguments,
    },
    filters={
        'positional_arguments': lambda x: x,
        'named_arguments': lambda x: x,
    },
)
class FixedCLISource(AbstractSource[ExpectedType]):
    def __init__(self, positional_arguments: Optional[List[str]] = None, named_arguments: Optional[List[str]] = None) -> None:
        if named_arguments is None and positional_arguments is None:
            raise ValueError("You need to pass a list of named arguments or a list of positional arguments, you haven't passed anything.")

        if named_arguments is not None:
            for parameter in named_arguments:
                if not parameter.isidentifier() or parameter == '_' or '__' in parameter:
                    raise ValueError(f'The "{parameter}" parameter is not valid for use on the command line (most likely, it is also not a valid Python name).')
        if positional_arguments is not None:
            for parameter in positional_arguments:
                if not parameter.isidentifier():
                    raise ValueError(f'The "{parameter}" parameter is not a valid Python identifier.')

        arguments_intersection = set([] if positional_arguments is None else positional_arguments) & set([] if named_arguments is None else named_arguments)
        if arguments_intersection:
            raise ValueError(f"The following parameters overlap among positional and named command line arguments: {', '.join(sorted(arguments_intersection))}")

        self.position_arguments = positional_arguments if positional_arguments is not None else []
        self.named_arguments = named_arguments if named_arguments is not None else []

        self.parser = ArgumentParser(add_help=False)

        for parameter in self.named_arguments:
            if len(parameter) == 1:
                full_parameter = f'-{parameter}'
            else:
                full_parameter = f'--{parameter.replace("_", "-")}'
            self.parser.add_argument(full_parameter, nargs='?', const=None, default=sentinel)

        for parameter in self.position_arguments:
            self.parser.add_argument(parameter, nargs='?', const=None, default=sentinel)

    def __getitem__(self, key: str) -> str:  # type: ignore[override]
        with open(os.devnull, 'w') as devnull, redirect_stderr(devnull):
            try:
                result: Union[str, InnerNoneType] = vars(self.parser.parse_args())[key]
                if result is sentinel:
                    raise KeyError(key)
                return result  # type: ignore[return-value]
            # TODO: from python 3.9 use exit_on_error
            except SystemExit as e:
                raise KeyError(key) from e  # pragma: no cover

    def type_awared_get(self, key: str, hint: Type[ExpectedType], default: ExpectedType = cast(ExpectedType, sentinel)) -> Optional[ExpectedType]:  # noqa: B008
        subresult = cast(Union[str, SentinelType], self.get(key, default))

        if hint is bool and key in self.named_arguments:
            if subresult is None:
                return cast(ExpectedType, True)
            if subresult is sentinel:
                return cast(ExpectedType, False)
            raise CLIFormatError("You can't pass values for boolean named fields to the CLI.")

        if subresult is default:
            if default is not sentinel:
                return default
            return None

        return from_string(cast(str, subresult), hint)

    @classmethod
    def for_library(cls, library_name: str) -> List['FixedCLISource[ExpectedType]']:
        if not library_name.isidentifier():
            raise ValueError('The library name can only be a valid Python identifier.')

        return []



from typing import List

from skelet import EnvSource, JSONSource, TOMLSource, YAMLSource
from skelet.sources.abstract import AbstractSource, ExpectedType


def for_tool(tool_name: str) -> List[AbstractSource[ExpectedType]]:
    return EnvSource.for_library(tool_name) + TOMLSource.for_library(tool_name) + YAMLSource.for_library(tool_name) + JSONSource.for_library(tool_name)  # type: ignore[return-value, operator]



import os
import platform
from functools import cached_property
from typing import Dict, List, Optional, Type, cast

from denial import InnerNoneType
from printo import repred
from simtypes import from_string

from skelet.errors import CaseError
from skelet.sources.abstract import AbstractSource, ExpectedType

sentinel = InnerNoneType()

@repred(filters={'prefix': lambda x: x != '', 'postfix': lambda x: x != '', 'case_sensitive': lambda x: x != False})  # type: ignore[arg-type]
class EnvSource(AbstractSource[ExpectedType]):
    def __init__(self, prefix: Optional[str] = '', postfix: Optional[str] = '', case_sensitive: bool = False) -> None:
        if platform.system() == 'Windows' and case_sensitive:
            raise OSError('On Windows, the environment variables are case-independent.')  # pragma: no cover

        self.prefix = prefix
        self.postfix = postfix
        self.case_sensitive = case_sensitive

    def __getitem__(self, key: str) -> str:  # type: ignore[override]
        full_key = f'{self.prefix}{key}{self.postfix}'
        if not self.case_sensitive:  # pragma: no cover
            full_key = full_key.upper()

        return self.data[full_key]

    @cached_property
    def data(self) -> Dict[str, str]:
        if self.case_sensitive:  # pragma: no cover
            return os.environ.copy()

        result = {}
        seen_keys: Dict[str, str] = {}
        for key, value in os.environ.items():
            capitalized_key = key.upper()
            if capitalized_key in seen_keys:
                if os.environ[key] != os.environ[seen_keys[capitalized_key]]:  # pragma: no cover
                    raise CaseError(f'There are 2 environment variables that are written the same way when capitalized: "{key}" and "{seen_keys[capitalized_key]}".')
            seen_keys[capitalized_key] = key
            result[capitalized_key] = value

        return result

    def type_awared_get(self, key: str, hint: Type[ExpectedType], default: ExpectedType = cast(ExpectedType, sentinel)) -> Optional[ExpectedType]:  # noqa: B008
        subresult = cast(str, self.get(key, default))

        if subresult is default:
            if default is not sentinel:
                return default
            return None

        return from_string(subresult, hint)

    @classmethod
    def for_library(cls, library_name: str) -> List['EnvSource[ExpectedType]']:
        if not library_name.isidentifier():
            raise ValueError('The library name can only be a valid Python identifier.')

        return [cls(prefix=f'{library_name}_'.upper())]



from typing import Any, Dict

from skelet.storage import Storage


def asdict(storage: Storage) -> Dict[str, Any]:
    if not isinstance(storage, Storage):
        raise TypeError('asdict() should be called on Storage class instances')

    result = {}

    for field_name, value in storage.__values__.items():
        result[field_name] = value

    return result
